from __future__ import annotations

from .base import PluginLoader


class Discovery(PluginLoader):
    """Discovery plugins."""


def get_discover(parser, args):
    discover_types = Discovery.entry_points_for("virtualenv.discovery")
    discovery_parser = parser.add_argument_group(
        title="discovery",
        description="discover and provide a target interpreter",
    )
    choices = _get_default_discovery(discover_types)
    # prefer the builtin if present, otherwise fallback to first defined type
    choices = sorted(choices, key=lambda a: 0 if a == "builtin" else 1)
    try:
        default_discovery = next(iter(choices))
    except StopIteration as e:
        msg = "No discovery plugin found. Try reinstalling virtualenv to fix this issue."
        raise RuntimeError(msg) from e
    discovery_parser.add_argument(
        "--discovery",
        choices=choices,
        default=default_discovery,
        required=False,
        help="interpreter discovery method",
    )
    options, _ = parser.parse_known_args(args)
    discover_class = discover_types[options.discovery]
    discover_class.add_parser_arguments(discovery_parser)
    options, _ = parser.parse_known_args(args, namespace=options)
    return discover_class(options)


def _get_default_discovery(discover_types):
    return list(discover_types.keys())


__all__ = [
    "Discovery",
    "get_discover",
]



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf import message as _message
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor
RESOURCE_REFERENCE_FIELD_NUMBER: _ClassVar[int]
resource_reference: _descriptor.FieldDescriptor
RESOURCE_DEFINITION_FIELD_NUMBER: _ClassVar[int]
resource_definition: _descriptor.FieldDescriptor
RESOURCE_FIELD_NUMBER: _ClassVar[int]
resource: _descriptor.FieldDescriptor

class ResourceDescriptor(_message.Message):
    __slots__ = (
        "type",
        "pattern",
        "name_field",
        "history",
        "plural",
        "singular",
        "style",
    )
    class History(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        HISTORY_UNSPECIFIED: _ClassVar[ResourceDescriptor.History]
        ORIGINALLY_SINGLE_PATTERN: _ClassVar[ResourceDescriptor.History]
        FUTURE_MULTI_PATTERN: _ClassVar[ResourceDescriptor.History]

    HISTORY_UNSPECIFIED: ResourceDescriptor.History
    ORIGINALLY_SINGLE_PATTERN: ResourceDescriptor.History
    FUTURE_MULTI_PATTERN: ResourceDescriptor.History
    class Style(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STYLE_UNSPECIFIED: _ClassVar[ResourceDescriptor.Style]
        DECLARATIVE_FRIENDLY: _ClassVar[ResourceDescriptor.Style]

    STYLE_UNSPECIFIED: ResourceDescriptor.Style
    DECLARATIVE_FRIENDLY: ResourceDescriptor.Style
    TYPE_FIELD_NUMBER: _ClassVar[int]
    PATTERN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_FIELD_NUMBER: _ClassVar[int]
    HISTORY_FIELD_NUMBER: _ClassVar[int]
    PLURAL_FIELD_NUMBER: _ClassVar[int]
    SINGULAR_FIELD_NUMBER: _ClassVar[int]
    STYLE_FIELD_NUMBER: _ClassVar[int]
    type: str
    pattern: _containers.RepeatedScalarFieldContainer[str]
    name_field: str
    history: ResourceDescriptor.History
    plural: str
    singular: str
    style: _containers.RepeatedScalarFieldContainer[ResourceDescriptor.Style]
    def __init__(
        self,
        type: _Optional[str] = ...,
        pattern: _Optional[_Iterable[str]] = ...,
        name_field: _Optional[str] = ...,
        history: _Optional[_Union[ResourceDescriptor.History, str]] = ...,
        plural: _Optional[str] = ...,
        singular: _Optional[str] = ...,
        style: _Optional[_Iterable[_Union[ResourceDescriptor.Style, str]]] = ...,
    ) -> None: ...

class ResourceReference(_message.Message):
    __slots__ = ("type", "child_type")
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CHILD_TYPE_FIELD_NUMBER: _ClassVar[int]
    type: str
    child_type: str
    def __init__(
        self, type: _Optional[str] = ..., child_type: _Optional[str] = ...
    ) -> None: ...



'''Partial class to handle Vultr Server (IPv4) API calls'''
from .utils import VultrBase, update_params


class VultrServerIPv4(VultrBase):
    '''Handles Vultr Server (IPv4) API calls'''
    def __init__(self, api_key):
        VultrBase.__init__(self, api_key)

    def create(self, subid, params=None):
        ''' /v1/server/create_ipv4
        POST - account
        Add a new IPv4 address to a server. You will start being billed for
        this immediately. The server will be rebooted unless you specify
        otherwise. You must reboot the server before the IPv4 address can be
        configured.

        Link: https://www.vultr.com/api/#server_create_ipv4
        '''
        params = update_params(params, {'SUBID': subid})
        return self.request('/v1/server/create_ipv4', params, 'POST')

    def destroy(self, subid, ipaddr, params=None):
        ''' /v1/server/destroy_ipv4
        POST - account
        Removes a secondary IPv4 address from a server. Your server will be
        hard-restarted. We suggest halting the machine gracefully before
        removing IPs.

        Link: https://www.vultr.com/api/#server_destroy_ipv4
        '''
        params = update_params(params, {
            'SUBID': subid,
            'ip': ipaddr
        })
        return self.request('/v1/server/destroy_ipv4', params, 'POST')

    def list(self, subid, params=None):
        ''' /v1/server/list_ipv4
        GET - account
        List the IPv4 information of a virtual machine. IP information is only
        available for virtual machines in the "active" state.

        Link: https://www.vultr.com/api/#server_list_ipv4
        '''
        params = update_params(params, {'SUBID': subid})
        return self.request('/v1/server/list_ipv4', params, 'GET')

    def reverse_default(self, subid, ipaddr, params=None):
        ''' /v1/server/reverse_default_ipv4
        POST - account
        Set a reverse DNS entry for an IPv4 address of a virtual
        machine to the original setting. Upon success, DNS changes
        may take 6-12 hours to become active.

        Link: https://www.vultr.com/api/#server_reverse_default_ipv4
        '''
        params = update_params(params, {
            'SUBID': subid,
            'ip': ipaddr
        })
        return self.request('/v1/server/reverse_default_ipv4', params, 'POST')

    def reverse_set(self, subid, ipaddr, entry, params=None):
        ''' /v1/server/reverse_set_ipv4
        POST - account
        Set a reverse DNS entry for an IPv4 address of a virtual machine. Upon
        success, DNS changes may take 6-12 hours to become active.

        Link: https://www.vultr.com/api/#server_reverse_set_ipv4
        '''
        params = update_params(params, {
            'SUBID': subid,
            'ip': ipaddr,
            'entry': entry
        })
        return self.request('/v1/server/reverse_set_ipv4', params, 'POST')



from __future__ import annotations

import sys
from collections import OrderedDict
from importlib.metadata import entry_points

importlib_metadata_version = ()


class PluginLoader:
    _OPTIONS = None
    _ENTRY_POINTS = None

    @classmethod
    def entry_points_for(cls, key):
        if sys.version_info >= (3, 10) or importlib_metadata_version >= (3, 6):
            return OrderedDict((e.name, e.load()) for e in cls.entry_points().select(group=key))
        return OrderedDict((e.name, e.load()) for e in cls.entry_points().get(key, {}))

    @staticmethod
    def entry_points():
        if PluginLoader._ENTRY_POINTS is None:
            PluginLoader._ENTRY_POINTS = entry_points()
        return PluginLoader._ENTRY_POINTS


class ComponentBuilder(PluginLoader):
    def __init__(self, interpreter, parser, name, possible) -> None:
        self.interpreter = interpreter
        self.name = name
        self._impl_class = None
        self.possible = possible
        self.parser = parser.add_argument_group(title=name)
        self.add_selector_arg_parse(name, list(self.possible))

    @classmethod
    def options(cls, key):
        if cls._OPTIONS is None:
            cls._OPTIONS = cls.entry_points_for(key)
        return cls._OPTIONS

    def add_selector_arg_parse(self, name, choices):
        raise NotImplementedError

    def handle_selected_arg_parse(self, options):
        selected = getattr(options, self.name)
        if selected not in self.possible:
            msg = f"No implementation for {self.interpreter}"
            raise RuntimeError(msg)
        self._impl_class = self.possible[selected]
        self.populate_selected_argparse(selected, options.app_data)
        return selected

    def populate_selected_argparse(self, selected, app_data):
        self.parser.description = f"options for {self.name} {selected}"
        self._impl_class.add_parser_arguments(self.parser, self.interpreter, app_data)

    def create(self, options):
        return self._impl_class(options, self.interpreter)


__all__ = [
    "ComponentBuilder",
    "PluginLoader",
]



'''Partial class to handle Vultr App API calls'''
from .utils import VultrBase


class VultrApp(VultrBase):
    '''Handles Vultr App API calls'''
    def __init__(self, api_key):
        VultrBase.__init__(self, api_key)

    def list(self, params=None):
        ''' /v1/app/list
        GET - account
        Retrieve a list of available applications. These
        refer to applications that can be launched when
        creating a Vultr VPS.

        Link: https://www.vultr.com/api/#app_app_list
        '''
        params = params if params else dict()
        return self.request('/v1/app/list', params, 'GET')



# Copyright 2022-2026 Broadcom.
# SPDX-License-Identifier: Apache-2.0
"""
Installation and finalization functions for the build process.
"""

from __future__ import annotations

import fnmatch
import hashlib
import json
import logging
import os
import os.path
import pathlib
import pprint
import re
import shutil
import sys
import tarfile
from typing import IO, TYPE_CHECKING

import relenv.relocate
from relenv.common import (
    LINUX,
    MODULE_DIR,
    MissingDependencyError,
    Version,
    download_url,
    format_shebang,
    runcmd,
)

if TYPE_CHECKING:
    from collections.abc import MutableMapping, Sequence
    from types import ModuleType

    from .builder import Dirs

# Type alias for path-like objects
PathLike = str | os.PathLike[str]

# Relenv PTH file content for bootstrapping
RELENV_PTH = (
    "import os; "
    "import sys; "
    "from importlib import util; "
    "from pathlib import Path; "
    "spec = util.spec_from_file_location("
    "'relenv.runtime', str(Path(__file__).parent / 'site-packages' / 'relenv' / 'runtime.py')"
    "); "
    "mod = util.module_from_spec(spec); "
    "sys.modules['relenv.runtime'] = mod; "
    "spec.loader.exec_module(mod); mod.bootstrap();"
)

log = logging.getLogger(__name__)


def patch_file(path: PathLike, old: str, new: str) -> None:
    """
    Search a file for a string to replace.

    :param path: Location of the file to search
    :type path: str
    :param old: The value that will be replaced (regex)
    :type old: str
    :param new: The value that will replace the 'old' value.
    :type path: str
    """
    log.debug("Patching file: %s", path)
    path = pathlib.Path(path)
    if not path.exists():
        log.warning("File not found for patching: %s", path)
        return

    content = path.read_text(encoding="utf-8")
    new_content = re.sub(old, new, content, flags=re.IGNORECASE | re.MULTILINE)
    path.write_text(new_content, encoding="utf-8")


def update_sbom_checksums(source_dir: PathLike, files_to_update: MutableMapping[str, PathLike]) -> None:
    """
    Update checksums in sbom.spdx.json for modified files.

    Python 3.12+ includes an SBOM (Software Bill of Materials) that tracks
    file checksums. When we update files (e.g., expat sources), we need to
    recalculate their checksums.

    :param source_dir: Path to the Python source directory
    :type source_dir: PathLike
    :param files_to_update: Mapping of SBOM relative paths to actual file paths
    :type files_to_update: MutableMapping[str, PathLike]
    """
    source_path = pathlib.Path(source_dir)
    spdx_json = source_path / "Misc" / "sbom.spdx.json"

    # SBOM only exists in Python 3.12+
    if not spdx_json.exists():
        log.debug("SBOM file not found, skipping checksum updates")
        return

    # Read the SBOM JSON
    with open(spdx_json) as f:
        data = json.load(f)

    # Compute checksums for each file
    checksums = {}
    for relative_path, file_path in files_to_update.items():
        file_path = pathlib.Path(file_path)
        if not file_path.exists():
            log.warning("File not found for checksum: %s", file_path)
            continue

        # Compute SHA1 and SHA256
        sha1 = hashlib.sha1()
        sha256 = hashlib.sha256()
        with open(file_path, "rb") as f:
            content = f.read()
            sha1.update(content)
            sha256.update(content)

        checksums[relative_path] = [
            {
                "algorithm": "SHA1",
                "checksumValue": sha1.hexdigest(),
            },
            {
                "algorithm": "SHA256",
                "checksumValue": sha256.hexdigest(),
            },
        ]
        log.debug(
            "Computed checksums for %s: SHA1=%s, SHA256=%s",
            relative_path,
            sha1.hexdigest(),
            sha256.hexdigest(),
        )

    # Update the SBOM with new checksums
    updated_count = 0
    for file_entry in data.get("files", []):
        file_name = file_entry.get("fileName")
        if file_name in checksums:
            file_entry["checksums"] = checksums[file_name]
            updated_count += 1
            log.info("Updated SBOM checksums for %s", file_name)

    # Write back the updated SBOM
    with open(spdx_json, "w") as f:
        json.dump(data, f, indent=2)

    log.info("Updated %d file checksums in SBOM", updated_count)


def patch_shebang(path: PathLike, old: str, new: str) -> bool:
    """
    Replace a file's shebang.

    :param path: The path of the file to patch
    :type path: str
    :param old: The old shebang, will only patch when this is found
    :type old: str
    :param name: The new shebang to be written
    :type name: str
    """
    with open(path, "rb") as fp:
        try:
            data = fp.read(len(old.encode())).decode()
        except UnicodeError:
            return False
        except Exception as exc:
            log.warning("Unhandled exception: %r", exc)
            return False
        if data != old:
            log.warning("Shebang doesn't match: %s %r != %r", path, old, data)
            return False
        data = fp.read().decode()
    with open(path, "w") as fp:
        fp.write(new)
        fp.write(data)
    with open(path) as fp:
        data = fp.read()
    log.info("Patched shebang of %s => %r", path, data)
    return True


def patch_shebangs(path: PathLike, old: str, new: str) -> None:
    """
    Traverse directory and patch shebangs.

    :param path: The of the directory to traverse
    :type path: str
    :param old: The old shebang, will only patch when this is found
    :type old: str
    :param name: The new shebang to be written
    :type name: str
    """
    for root, _dirs, files in os.walk(str(path)):
        for file in files:
            patch_shebang(os.path.join(root, file), old, new)


def _load_sysconfigdata_template() -> str:
    """Load the sysconfigdata template from disk.

    Returns:
        The Python code template for sysconfigdata module.

    Note:
        This is loaded from a .py file rather than embedded as a string
        to enable syntax checking, IDE support, and easier maintenance.
        Follows CPython convention of separating data from code.
    """
    template_path = pathlib.Path(__file__).parent / "_sysconfigdata_template.py"
    template_content = template_path.read_text(encoding="utf-8")

    # Extract only the code after the docstring
    # Skip the copyright header and module docstring
    lines = template_content.split("\n")
    code_lines = []
    found_code = False

    for line in lines:
        # Skip until we find the first import statement
        if not found_code:
            if line.startswith("import ") or line.startswith("from "):
                found_code = True
            else:
                continue

        code_lines.append(line)

    return "\n".join(code_lines)


def update_ensurepip(directory: pathlib.Path) -> None:
    """
    Update bundled dependencies for ensurepip (pip & setuptools).
    """
    # ensurepip bundle location
    bundle_dir = directory / "ensurepip" / "_bundled"

    # Make sure the destination directory exists
    bundle_dir.mkdir(parents=True, exist_ok=True)

    # Detect existing whl. Later versions of python don't include setuptools. We
    # only want to update whl files that python expects to be there
    pip_version = "25.2"
    setuptools_version = "80.9.0"
    update_pip = False
    update_setuptools = False
    for file in bundle_dir.glob("*.whl"):
        log.debug("Checking whl: %s", str(file))
        if file.name.startswith("pip-"):
            found_version = file.name.split("-")[1]
            log.debug("Found version %s", found_version)
            if Version(found_version) >= Version(pip_version):
                log.debug("Found correct pip version or newer: %s", found_version)
            else:
                file.unlink()
                update_pip = True
        if file.name.startswith("setuptools-"):
            found_version = file.name.split("-")[1]
            log.debug("Found version %s", found_version)
            if Version(found_version) >= Version(setuptools_version):
                log.debug("Found correct setuptools version or newer: %s", found_version)
            else:
                file.unlink()
                update_setuptools = True

    # Download whl files and update __init__.py
    init_file = directory / "ensurepip" / "__init__.py"
    if update_pip:
        whl = f"pip-{pip_version}-py3-none-any.whl"
        whl_path = "b7/3f/945ef7ab14dc4f9d7f40288d2df998d1837ee0888ec3659c813487572faa"
        url = f"https://files.pythonhosted.org/packages/{whl_path}/{whl}"
        download_url(url=url, dest=bundle_dir)
        assert (bundle_dir / whl).exists()

        # Update __init__.py
        old = "^_PIP_VERSION.*"
        new = f'_PIP_VERSION = "{pip_version}"'
        patch_file(path=init_file, old=old, new=new)

    # setuptools
    if update_setuptools:
        whl = f"setuptools-{setuptools_version}-py3-none-any.whl"
        whl_path = "a3/dc/17031897dae0efacfea57dfd3a82fdd2a2aeb58e0ff71b77b87e44edc772"
        url = f"https://files.pythonhosted.org/packages/{whl_path}/{whl}"
        download_url(url=url, dest=bundle_dir)
        assert (bundle_dir / whl).exists()

        # setuptools
        old = "^_SETUPTOOLS_VERSION.*"
        new = f'_SETUPTOOLS_VERSION = "{setuptools_version}"'
        patch_file(path=init_file, old=old, new=new)

    log.debug("ensurepip __init__.py contents:")
    log.debug(init_file.read_text())


def install_sysdata(
    mod: ModuleType,
    destfile: PathLike,
    buildroot: PathLike,
    toolchain: PathLike | None,
) -> None:
    """
    Create a Relenv Python environment's sysconfigdata.

    Helper method used by the `finalize` build method to create a Relenv
    Python environment's sysconfigdata.

    :param mod: The module to operate on
    :type mod: ``types.ModuleType``
    :param destfile: Path to the file to write the data to
    :type destfile: str
    :param buildroot: Path to the root of the build
    :type buildroot: str
    :param toolchain: Path to the root of the toolchain
    :type toolchain: str
    """
    data = {}

    def fbuildroot(s: str) -> str:
        return s.replace(str(buildroot), "{BUILDROOT}")

    def ftoolchain(s: str) -> str:
        return s.replace(str(toolchain), "{TOOLCHAIN}")

    # XXX: keymap is not used, remove it?
    # keymap = {
    #    "BINDIR": (fbuildroot,),
    #    "BINLIBDEST": (fbuildroot,),
    #    "CFLAGS": (fbuildroot, ftoolchain),
    #    "CPPLAGS": (fbuildroot, ftoolchain),
    #    "CXXFLAGS": (fbuildroot, ftoolchain),
    #    "datarootdir": (fbuildroot,),
    #    "exec_prefix": (fbuildroot,),
    #    "LDFLAGS": (fbuildroot, ftoolchain),
    #    "LDSHARED": (fbuildroot, ftoolchain),
    #    "LIBDEST": (fbuildroot,),
    #    "prefix": (fbuildroot,),
    #    "SCRIPTDIR": (fbuildroot,),
    # }
    for key in sorted(mod.build_time_vars):
        val = mod.build_time_vars[key]
        if isinstance(val, str):
            for _ in (fbuildroot, ftoolchain):
                val = _(val)
                log.info("SYSCONFIG [%s] %s => %s", key, mod.build_time_vars[key], val)
        data[key] = val

    sysconfigdata_code = _load_sysconfigdata_template()
    with open(destfile, "w", encoding="utf8") as f:
        f.write("# system configuration generated and used by the relenv at runtime\n")
        f.write("_build_time_vars = ")
        pprint.pprint(data, stream=f)
        f.write(sysconfigdata_code)


def find_sysconfigdata(pymodules: PathLike) -> str:
    """
    Find sysconfigdata directory for python installation.

    :param pymodules: Path to python modules (e.g. lib/python3.10)
    :type pymodules: str

    :return: The name of the sysconig data module
    :rtype: str
    """
    for root, dirs, files in os.walk(pymodules):
        for file in files:
            if file.find("sysconfigdata") > -1 and file.endswith(".py"):
                return file[:-3]
    raise MissingDependencyError("Unable to locate sysconfigdata module")


def install_runtime(sitepackages: PathLike) -> None:
    """
    Install a base relenv runtime.
    """
    site_dir = pathlib.Path(sitepackages)
    relenv_pth = site_dir / "relenv.pth"
    with open(str(relenv_pth), "w") as fp:
        fp.write(RELENV_PTH)

    # Lay down relenv.runtime, we'll pip install the rest later
    relenv = site_dir / "relenv"
    os.makedirs(relenv, exist_ok=True)

    for name in [
        "runtime.py",
        "relocate.py",
        "common.py",
        "buildenv.py",
        "__init__.py",
    ]:
        src = MODULE_DIR / name
        dest = relenv / name
        with open(src) as rfp:
            with open(dest, "w") as wfp:
                wfp.write(rfp.read())


def finalize(
    env: MutableMapping[str, str],
    dirs: Dirs,
    logfp: IO[str],
) -> None:
    """
    Run after we've fully built python.

    This method enhances the newly created python with Relenv's runtime hacks.

    :param env: The environment dictionary
    :type env: dict
    :param dirs: The working directories
    :type dirs: ``relenv.build.common.Dirs``
    :param logfp: A handle for the log file
    :type logfp: file
    """
    # Run relok8 to make sure the rpaths are relocatable.
    # Modules that don't link to relenv libs will have their RPATH removed
    relenv.relocate.main(
        dirs.prefix,
        log_file_name=str(dirs.logs / "relocate.py.log"),
    )
    # Install relenv-sysconfigdata module
    libdir = pathlib.Path(dirs.prefix) / "lib"

    def find_pythonlib(libdir: pathlib.Path) -> str | None:
        for _root, dirs, _files in os.walk(libdir):
            for entry in dirs:
                if entry.startswith("python"):
                    return entry
        return None

    python_lib = find_pythonlib(libdir)
    if python_lib is None:
        raise MissingDependencyError("Unable to locate python library directory")

    pymodules = libdir / python_lib

    # update ensurepip
    update_ensurepip(pymodules)

    cwd = os.getcwd()
    modname = find_sysconfigdata(pymodules)
    path = sys.path
    sys.path = [str(pymodules)]
    try:
        mod = __import__(str(modname))
    finally:
        os.chdir(cwd)
        sys.path = path

    dest = pymodules / f"{modname}.py"
    install_sysdata(mod, dest, dirs.prefix, dirs.toolchain)

    # Lay down site customize
    bindir = pathlib.Path(dirs.prefix) / "bin"
    sitepackages = pymodules / "site-packages"
    install_runtime(sitepackages)

    # Install pip
    python_exe = str(dirs.prefix / "bin" / "python3")
    if env["RELENV_HOST_ARCH"] != env["RELENV_BUILD_ARCH"]:
        env["RELENV_CROSS"] = str(dirs.prefix)
        python_exe = env["RELENV_NATIVE_PY"]
    logfp.write("\nRUN ENSURE PIP\n")

    env.pop("RELENV_BUILDENV")

    runcmd(
        [python_exe, "-m", "ensurepip"],
        env=env,
        stderr=logfp,
        stdout=logfp,
    )

    # Fix the shebangs in the scripts python layed down. Order matters.
    shebangs = [
        "#!{}".format(bindir / f"python{env['RELENV_PY_MAJOR_VERSION']}"),
        "#!{}".format(bindir / f"python{env['RELENV_PY_MAJOR_VERSION'].split('.', 1)[0]}"),
    ]
    newshebang = format_shebang("/python3")
    for shebang in shebangs:
        log.info("Patch shebang %r with  %r", shebang, newshebang)
        patch_shebangs(
            str(pathlib.Path(dirs.prefix) / "bin"),
            shebang,
            newshebang,
        )

    if sys.platform == "linux":
        pyconf = f"config-{env['RELENV_PY_MAJOR_VERSION']}-{env['RELENV_HOST']}"
        patch_shebang(
            str(pymodules / pyconf / "python-config.py"),
            "#!{}".format(str(bindir / f"python{env['RELENV_PY_MAJOR_VERSION']}")),
            format_shebang("../../../bin/python3"),
        )

        toolchain_path = dirs.toolchain
        if toolchain_path is None:
            raise MissingDependencyError("Toolchain path is required for linux builds")
        shutil.copy(
            pathlib.Path(toolchain_path) / env["RELENV_HOST"] / "sysroot" / "lib" / "libstdc++.so.6",
            libdir,
        )

    # Moved in python 3.13 or removed?
    if (pymodules / "cgi.py").exists():
        patch_shebang(
            str(pymodules / "cgi.py"),
            "#! /usr/local/bin/python",
            format_shebang("../../bin/python3"),
        )

    def runpip(pkg: str | os.PathLike[str], upgrade: bool = False) -> None:
        logfp.write(f"\nRUN PIP {pkg} {upgrade}\n")
        target: pathlib.Path | None = None
        python_exe = str(dirs.prefix / "bin" / "python3")
        if sys.platform == LINUX:
            if env["RELENV_HOST_ARCH"] != env["RELENV_BUILD_ARCH"]:
                target = pymodules / "site-packages"
                python_exe = env["RELENV_NATIVE_PY"]
        cmd = [
            python_exe,
            "-m",
            "pip",
            "install",
            str(pkg),
        ]
        if upgrade:
            cmd.append("--upgrade")
        if target:
            cmd.append(f"--target={target}")
        runcmd(cmd, env=env, stderr=logfp, stdout=logfp)

    runpip("wheel")
    # This needs to handle running from the root of the git repo and also from
    # an installed Relenv
    if (MODULE_DIR.parent / ".git").exists():
        runpip(MODULE_DIR.parent, upgrade=True)
    else:
        runpip("relenv", upgrade=True)
    globs = [
        "/bin/python*",
        "/bin/pip*",
        "/bin/relenv",
        "/lib/python*/ensurepip/*",
        "/lib/python*/site-packages/*",
        "/include/*",
        "*.so",
        "/lib/*.so.*",
        "*.py",
        # Mac specific, factor this out
        "*.dylib",
    ]
    archive = f"{dirs.prefix}.tar.xz"
    log.info("Archive is %s", archive)
    with tarfile.open(archive, mode="w:xz") as fp:
        create_archive(fp, dirs.prefix, globs, logfp)


def create_archive(
    tarfp: tarfile.TarFile,
    toarchive: PathLike,
    globs: Sequence[str],
    logfp: IO[str] | None = None,
) -> None:
    """
    Create an archive.

    :param tarfp: A pointer to the archive to be created
    :type tarfp: file
    :param toarchive: The path to the directory to archive
    :type toarchive: str
    :param globs: A list of filtering patterns to match against files to be added
    :type globs: list
    :param logfp: A pointer to the log file
    :type logfp: file
    """
    log.debug("Current directory %s", os.getcwd())
    log.debug("Creating archive %s", tarfp.name)
    for root, _dirs, files in os.walk(toarchive):
        relroot = pathlib.Path(root).relative_to(toarchive)
        for f in files:
            relpath = relroot / f
            matches = False
            for g in globs:
                candidate = pathlib.Path("/") / relpath
                if fnmatch.fnmatch(str(candidate), g):
                    matches = True
                    break
            if matches:
                log.debug("Adding %s", relpath)
                tarfp.add(relpath, arcname=str(relpath), recursive=False)
            else:
                log.debug("Skipping %s", relpath)



"""
NAPALM Route
============

Retrieves route details from network devices.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------
- :mod:`NAPALM proxy minion <salt.proxy.napalm>`

.. versionadded:: 2016.11.0
"""

import logging

# import NAPALM utils
import salt.utils.napalm
from salt.utils.napalm import proxy_napalm_wrap

log = logging.getLogger(__file__)


# ----------------------------------------------------------------------------------------------------------------------
# module properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "route"
__proxyenabled__ = ["napalm"]
# uses NAPALM-based proxy to interact with network devices
__virtual_aliases__ = ("napalm_route",)

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


@proxy_napalm_wrap
def show(destination, protocol=None, **kwargs):  # pylint: disable=unused-argument
    """
    Displays all details for a certain route learned via a specific protocol.
    If the protocol is not specified, will return all possible routes.

    .. note::

        This function return the routes from the RIB.
        In case the destination prefix is too short,
        there may be too many routes matched.
        Therefore in cases of devices having a very high number of routes
        it may be necessary to adjust the prefix length and request
        using a longer prefix.

    destination
        destination prefix.

    protocol (optional)
        protocol used to learn the routes to the destination.

    .. versionchanged:: 2017.7.0

    CLI Example:

    .. code-block:: bash

        salt 'my_router' route.show 172.16.0.0/25
        salt 'my_router' route.show 172.16.0.0/25 bgp

    Output example:

    .. code-block:: python

        {
            '172.16.0.0/25': [
                {
                    'protocol': 'BGP',
                    'last_active': True,
                    'current_active': True,
                    'age': 1178693,
                    'routing_table': 'inet.0',
                    'next_hop': '192.168.0.11',
                    'outgoing_interface': 'xe-1/1/1.100',
                    'preference': 170,
                    'selected_next_hop': False,
                    'protocol_attributes': {
                        'remote_as': 65001,
                        'metric': 5,
                        'local_as': 13335,
                        'as_path': '',
                        'remote_address': '192.168.0.11',
                        'metric2': 0,
                        'local_preference': 0,
                        'communities': [
                            '0:2',
                            'no-export'
                        ],
                        'preference2': -1
                    },
                    'inactive_reason': ''
                },
                {
                    'protocol': 'BGP',
                    'last_active': False,
                    'current_active': False,
                    'age': 2359429,
                    'routing_table': 'inet.0',
                    'next_hop': '192.168.0.17',
                    'outgoing_interface': 'xe-1/1/1.100',
                    'preference': 170,
                    'selected_next_hop': True,
                    'protocol_attributes': {
                        'remote_as': 65001,
                        'metric': 5,
                        'local_as': 13335,
                        'as_path': '',
                        'remote_address': '192.168.0.17',
                        'metric2': 0,
                        'local_preference': 0,
                        'communities': [
                            '0:3',
                            'no-export'
                        ],
                        'preference2': -1
                    },
                    'inactive_reason': 'Not Best in its group - Router ID'
                }
            ]
        }
    """

    return salt.utils.napalm.call(
        napalm_device,  # pylint: disable=undefined-variable
        "get_route_to",
        **{"destination": destination, "protocol": protocol},
    )



"""
Backward compatible shim layer for pki interaction

.. versionadded:: 3008.0

The ``localfs_key`` is a shim driver meant to allow the salt.cache
subsystem to interact with the existing master pki folder/file structure
without any migration from previous versions of salt.  It is not meant for
general purpose use and should not be used outside of the master auth system.

The main difference from before is the 'state' of the key, ie accepted/rejected
is now stored in the data itself, as opposed to the cache equivalent of a bank
previously.

store and fetch handle ETL from new style, where data itself contains key
state, to old style, where folder and/or bank contain state.
flush/list/contains/updated are left as nearly equivalent to localfs, without
the .p file extension to work with legacy keys via banks.
"""

import errno
import logging
import os
import os.path

try:
    import pwd
except ImportError:
    pwd = None
import shutil
import stat
import tempfile
from pathlib import Path

import salt.utils.atomicfile
import salt.utils.files
import salt.utils.stringutils
from salt.exceptions import SaltCacheError
from salt.utils.verify import clean_path, valid_id

log = logging.getLogger(__name__)

__func_alias__ = {"list_": "list"}


BASE_MAPPING = {
    "minions_pre": "pending",
    "minions_rejected": "rejected",
    "minions": "accepted",
    "minions_denied": "denied",
}

# master_keys keys that if fetched, even with cluster_id set, will still refer
# to pki_dir instead of cluster_pki_dir
NON_CLUSTERED_MASTER_KEYS = []


# we explicitly override cache dir to point to pki here
def init_kwargs(kwargs):
    """
    setup kwargs for cache functions
    """
    if __opts__["__role"] != "minion":
        global NON_CLUSTERED_MASTER_KEYS
        NON_CLUSTERED_MASTER_KEYS = [
            "master.pem",
            "master.pub",
            f"{__opts__['master_sign_key_name']}.pem",
            f"{__opts__['master_sign_key_name']}.pub",
            f"{__opts__['id'].removesuffix('_master')}.pub",
            f"{__opts__['id'].removesuffix('_master')}.pem",
            __opts__.get(
                "master_pubkey_signature", f"{__opts__['id']}_pubkey_signature"
            ),
        ]

    if "pki_dir" in kwargs:
        pki_dir = kwargs["pki_dir"]
    elif __opts__.get("cluster_id"):
        pki_dir = __opts__["cluster_pki_dir"]
    else:
        pki_dir = __opts__["pki_dir"]

    user = kwargs.get("user", __opts__.get("user"))

    return {"cachedir": pki_dir, "user": user}


def store(bank, key, data, cachedir, user, **kwargs):
    """
    Store key state information. storing a accepted/pending/rejected state
    means clearing it from the other 2. denied is handled separately
    """
    base = None
    if bank in ["keys", "denied_keys"] and not valid_id(__opts__, key):
        raise SaltCacheError(f"key {key} is not a valid minion_id")

    if bank not in ["keys", "denied_keys", "master_keys"]:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    if __opts__["permissive_pki_access"]:
        umask = 0o0700
    else:
        umask = 0o0750

    if bank == "keys":
        if data["state"] == "rejected":
            base = "minions_rejected"
        elif data["state"] == "pending":
            base = "minions_pre"
        elif data["state"] == "accepted":
            base = "minions"
        else:
            raise SaltCacheError("Unrecognized data/bank: {}".format(data["state"]))
        data = data["pub"]
    elif bank == "denied_keys":
        # denied keys is a list post migration, but is a single key in legacy
        data = data[0]
        base = "minions_denied"
    elif bank == "master_keys":
        # private keys are separate from permissive_pki_access
        umask = 0o277
        base = ""
        # even in clustered mode, master and signing keys live in the
        # non-clustered pki dir
        if key in NON_CLUSTERED_MASTER_KEYS:
            cachedir = __opts__["pki_dir"]

    savefn = Path(cachedir) / base / key
    base_dir = savefn.parent

    if not clean_path(cachedir, str(savefn), subdir=True):
        raise SaltCacheError(f"key {key} is not a valid key path.")

    try:
        os.makedirs(base_dir)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise SaltCacheError(
                f"The cache directory, {base_dir}, could not be created: {exc}"
            )

    # delete current state before re-serializing new state
    flush(bank, key, cachedir, **kwargs)

    tmpfh, tmpfname = tempfile.mkstemp(dir=base_dir)
    os.close(tmpfh)

    if user and not salt.utils.platform.is_windows():
        try:
            uid = pwd.getpwnam(user).pw_uid
            os.chown(tmpfname, uid, -1)
        except (KeyError, ImportError, OSError, NameError):
            # The specified user was not found, allow the backup systems to
            # report the error
            pass

    try:
        with salt.utils.files.set_umask(umask):
            with salt.utils.files.fopen(tmpfname, "w+b") as fh_:
                fh_.write(salt.utils.stringutils.to_bytes(data))

            if bank == "master_keys":
                os.chmod(tmpfname, 0o400)

        # On Windows, os.rename will fail if the destination file exists.
        salt.utils.atomicfile.atomic_rename(tmpfname, savefn)
    except OSError as exc:
        raise SaltCacheError(
            f"There was an error writing the cache file, base={base}: {exc}"
        )


def fetch(bank, key, cachedir, **kwargs):
    """
    Fetch and construct state data for a given minion based on the bank and id
    """
    if bank in ["keys", "denied_keys"] and not valid_id(__opts__, key):
        raise SaltCacheError(f"key {key} is not a valid minion_id")

    if bank not in ["keys", "denied_keys", "master_keys"]:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    if not clean_path(cachedir, key, subdir=True):
        raise SaltCacheError(f"key {key} is not a valid key path.")

    if key == ".key_cache":
        raise SaltCacheError("trying to read key_cache, there is a bug at call-site")
    try:
        if bank == "keys":
            for state, bank in [
                ("rejected", "minions_rejected"),
                ("pending", "minions_pre"),
                ("accepted", "minions"),
            ]:
                keyfile = Path(cachedir, bank, key)

                if not clean_path(cachedir, str(keyfile), subdir=True):
                    raise SaltCacheError(f"key {key} is not a valid key path.")

                if keyfile.is_file() and not keyfile.is_symlink():
                    with salt.utils.files.fopen(keyfile, "r") as fh_:
                        return {"state": state, "pub": fh_.read()}
            return None
        elif bank == "denied_keys":
            # there can be many denied keys per minion post refactor, but only 1
            # with the filesystem, so return a list of 1
            pubfn_denied = os.path.join(cachedir, "minions_denied", key)

            if not clean_path(cachedir, pubfn_denied, subdir=True):
                raise SaltCacheError(f"key {key} is not a valid key path.")

            if os.path.isfile(pubfn_denied):
                with salt.utils.files.fopen(pubfn_denied, "r") as fh_:
                    return [fh_.read()]
        elif bank == "master_keys":
            if key in NON_CLUSTERED_MASTER_KEYS:
                cachedir = __opts__["pki_dir"]

            keyfile = Path(cachedir, key)

            if not clean_path(cachedir, str(keyfile), subdir=True):
                raise SaltCacheError(f"key {key} is not a valid key path.")

            if keyfile.is_file() and not keyfile.is_symlink():
                with salt.utils.files.fopen(keyfile, "r") as fh_:
                    return fh_.read()
        else:
            raise SaltCacheError(f'unrecognized bank "{bank}"')
    except OSError as exc:
        raise SaltCacheError(
            'There was an error reading the cache bank "{}", key "{}": {}'.format(
                bank, key, exc
            )
        )


def updated(bank, key, cachedir, **kwargs):
    """
    Return the epoch of the mtime for this cache file
    """
    if not valid_id(__opts__, key):
        raise SaltCacheError(f"key {key} is not a valid minion_id")

    if bank == "keys":
        bases = [base for base in BASE_MAPPING if base != "minions_denied"]
    elif bank == "denied_keys":
        bases = ["minions_denied"]
    elif bank == "master_keys":
        if key in NON_CLUSTERED_MASTER_KEYS:
            cachedir = __opts__["pki_dir"]
        bases = [""]
    else:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    for dir in bases:
        keyfile = Path(cachedir, dir, key)

        if not clean_path(cachedir, str(keyfile), subdir=True):
            raise SaltCacheError(f"key {key} is not a valid key path.")

        if keyfile.is_file() and not keyfile.is_symlink():
            try:
                return int(os.path.getmtime(keyfile))
            except OSError as exc:
                raise SaltCacheError(
                    'There was an error reading the mtime for "{}": {}'.format(
                        keyfile, exc
                    )
                )
    log.debug('pki file "%s" does not exist in accepted/rejected/pending', key)
    return


def flush(bank, key=None, cachedir=None, **kwargs):
    """
    Remove the key from the cache bank with all the key content.
    flush can take a legacy bank or a keys/denied_keys modern bank
    """
    if bank in ["keys", "denied_keys"] and not valid_id(__opts__, key):
        raise SaltCacheError(f"key {key} is not a valid minion_id")

    if cachedir is None:
        raise SaltCacheError("cachedir missing")

    if bank == "keys":
        bases = [base for base in BASE_MAPPING if base != "minions_denied"]
    elif bank == "denied_keys":
        bases = ["minions_denied"]
    elif bank == "master_keys":
        if key in NON_CLUSTERED_MASTER_KEYS:
            cachedir = __opts__["pki_dir"]
        bases = [""]
    else:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    flushed = False

    for base in bases:
        try:
            if key is None:
                target = os.path.join(cachedir, base)
                if not os.path.isdir(target):
                    return False
                shutil.rmtree(target)
            else:
                target = os.path.join(cachedir, base, key)

                if not clean_path(cachedir, target, subdir=True):
                    raise SaltCacheError(f"key {key} is not a valid key path.")

                if not os.path.isfile(target):
                    continue

                # necessary on windows, otherwise PermissionError: [WinError 5] Access is denied
                os.chmod(target, stat.S_IWRITE)

                os.remove(target)
                flushed = True
        except OSError as exc:
            if exc.errno != errno.ENOENT:
                raise SaltCacheError(f'There was an error removing "{target}": {exc}')

    return flushed


def list_(bank, cachedir, **kwargs):
    """
    Return an iterable object containing all entries stored in the specified bank.
    Uses internal mmap index for O(1) performance when available.
    """
    if bank == "keys":
        bases = [base for base in BASE_MAPPING if base != "minions_denied"]
    elif bank == "denied_keys":
        bases = ["minions_denied"]
    elif bank == "master_keys":
        bases = [""]
    else:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    ret = []
    for base in bases:
        base = os.path.join(cachedir, os.path.normpath(base))
        if not os.path.isdir(base):
            continue
        try:
            items = os.listdir(base)
        except OSError as exc:
            raise SaltCacheError(
                f'There was an error accessing directory "{base}": {exc}'
            )
        for item in items:
            # salt foolishly dumps a file here for key cache, ignore it
            if item == ".key_cache":
                continue

            keyfile = Path(cachedir, base, item)

            if (
                bank in ["keys", "denied_keys"] and not valid_id(__opts__, item)
            ) or not clean_path(cachedir, str(keyfile), subdir=True):
                log.error("saw invalid id %s, discarding", item)
                continue

            if keyfile.is_file() and not keyfile.is_symlink():
                ret.append(item)
    return ret


def list_all(bank, cachedir, include_data=False, **kwargs):
    """
    Return all entries with their data from the specified bank.
    This is much faster than calling list() + fetch() for each item.
    Returns a dict of {key: data}.

    If include_data is False (default), only the state is returned for 'keys' bank,
    avoiding expensive file reads.
    """
    if bank not in ["keys", "denied_keys"]:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    ret = {}

    if bank == "keys":
        # Map directory names to states
        state_mapping = {
            "minions": "accepted",
            "minions_pre": "pending",
            "minions_rejected": "rejected",
        }

        for dir_name, state in state_mapping.items():
            dir_path = os.path.join(cachedir, dir_name)
            if not os.path.isdir(dir_path):
                continue

            try:
                with os.scandir(dir_path) as it:
                    for entry in it:
                        if not entry.is_file() or entry.is_symlink():
                            continue
                        if entry.name.startswith("."):
                            continue
                        if not valid_id(__opts__, entry.name):
                            continue
                        if not clean_path(cachedir, entry.path, subdir=True):
                            continue

                        if include_data:

                            # Read the public key
                            try:
                                with salt.utils.files.fopen(entry.path, "r") as fh_:
                                    pub_key = fh_.read()
                                ret[entry.name] = {"state": state, "pub": pub_key}
                            except OSError as exc:
                                log.error(
                                    "Error reading key file %s: %s", entry.path, exc
                                )
                        else:
                            # Just return the state, no disk read
                            ret[entry.name] = {"state": state}
            except OSError as exc:
                log.error("Error scanning directory %s: %s", dir_path, exc)

    elif bank == "denied_keys":
        # Denied keys work differently - multiple keys per minion ID
        dir_path = os.path.join(cachedir, "minions_denied")
        if os.path.isdir(dir_path):
            try:
                with os.scandir(dir_path) as it:
                    for entry in it:
                        if not entry.is_file() or entry.is_symlink():
                            continue
                        if not valid_id(__opts__, entry.name):
                            continue
                        if not clean_path(cachedir, entry.path, subdir=True):
                            continue

                        try:
                            with salt.utils.files.fopen(entry.path, "r") as fh_:
                                ret[entry.name] = fh_.read()
                        except OSError as exc:
                            log.error(
                                "Error reading denied key %s: %s", entry.path, exc
                            )
            except OSError as exc:
                log.error("Error scanning denied keys directory: %s", exc)

    return ret


def contains(bank, key, cachedir, **kwargs):
    """
    Checks if the specified bank contains the specified key.
    Uses internal mmap index for O(1) performance when available.
    """
    if bank in ["keys", "denied_keys"] and not valid_id(__opts__, key):
        raise SaltCacheError(f"key {key} is not a valid minion_id")

    if bank == "keys":
        bases = [base for base in BASE_MAPPING if base != "minions_denied"]
    elif bank == "denied_keys":
        bases = ["minions_denied"]
    elif bank == "master_keys":
        if key in NON_CLUSTERED_MASTER_KEYS:
            cachedir = __opts__["pki_dir"]
        bases = [""]
    else:
        raise SaltCacheError(f"Unrecognized bank: {bank}")

    for base in bases:
        keyfile = Path(cachedir, base, key)

        if not clean_path(cachedir, str(keyfile), subdir=True):
            raise SaltCacheError(f"key {key} is not a valid key path.")

        if keyfile.is_file() and not keyfile.is_symlink():
            return True

    return False



"""
Helpers for replicating ``file_roots`` and ``pillar_roots`` between
cluster masters that don't share a filesystem.

The cluster join-reply embeds the responder's current local state-tree
contents alongside the keys dump.  A late-joining master applies the
contents into its own local roots before becoming a Raft learner so it
can serve states/pillars it would otherwise know nothing about.

Used by :mod:`salt.channel.server` for the join handshake and (later)
by a ``cluster.sync_roots`` runner for ad-hoc updates after roots are
edited on a peer.
"""

import logging
import os
from pathlib import Path

import salt.utils.files

log = logging.getLogger(__name__)

# File and directory names skipped when collecting a roots tree.
# These are version-control artefacts and editor-temporary files that
# we never want to ship to peers.
_SKIP_DIR_NAMES = frozenset({".git", ".hg", ".svn", "__pycache__", ".tox"})


def _is_skipped_path(rel_parts):
    """Return ``True`` if any path component is on the skip list."""
    return any(part in _SKIP_DIR_NAMES for part in rel_parts)


def collect_root_tree(roots_map):
    """
    Build a wire-friendly snapshot of a ``file_roots``/``pillar_roots``
    mapping.

    :param roots_map: ``{env: [path, path, ...]}`` from ``opts``.
    :return: ``{env: [{"path": rel, "mode": int, "data": bytes}, ...]}``.

    Only regular files are included; symlinks, sockets and unreadable
    entries are skipped.  Multiple roots for the same env are flattened
    in declaration order — earlier roots win on path conflicts.
    """
    out = {}
    for env, paths in (roots_map or {}).items():
        files = []
        seen = set()
        for root in paths or []:
            root_p = Path(root)
            if not root_p.is_dir():
                continue
            for sub in root_p.rglob("*"):
                try:
                    if sub.is_symlink() or not sub.is_file():
                        continue
                    rel_parts = sub.relative_to(root_p).parts
                except (OSError, ValueError):
                    continue
                if not rel_parts or _is_skipped_path(rel_parts):
                    continue
                rel = "/".join(rel_parts)
                if rel in seen:
                    continue
                try:
                    data = sub.read_bytes()
                    mode = sub.stat().st_mode & 0o777
                except OSError as exc:
                    log.warning("file_sync: skipping unreadable %s: %s", sub, exc)
                    continue
                files.append({"path": rel, "mode": mode, "data": data})
                seen.add(rel)
        if files:
            out[env] = files
    return out


def apply_root_tree(roots_map, dump):
    """
    Materialise *dump* (from :func:`collect_root_tree`) under the local
    ``roots_map``.

    :param roots_map: ``{env: [path, path, ...]}`` from ``opts``.  Files
        for env *e* are written under ``roots_map[e][0]``; envs that are
        absent or empty are skipped.
    :param dump: the snapshot dict to apply.
    :return: number of files written.
    """
    written = 0
    for env, files in (dump or {}).items():
        roots = (roots_map or {}).get(env)
        if not roots:
            log.debug(
                "file_sync: env %r not configured locally; skipping %d files",
                env,
                len(files),
            )
            continue
        target = Path(roots[0])
        target.mkdir(parents=True, exist_ok=True)
        for entry in files:
            rel = entry.get("path") if isinstance(entry, dict) else None
            data = entry.get("data") if isinstance(entry, dict) else None
            mode = entry.get("mode", 0o644) if isinstance(entry, dict) else 0o644
            if not rel or data is None:
                continue
            # msgpack round-trip via salt.payload may turn bytes back into
            # str; coerce so binary files round-trip cleanly.
            if isinstance(data, str):
                data = data.encode("utf-8", errors="surrogateescape")
            dst = target / rel
            try:
                dst.parent.mkdir(parents=True, exist_ok=True)
                with salt.utils.files.fopen(dst, "wb") as fp:
                    fp.write(data)
                os.chmod(dst, mode)
                written += 1
            except OSError as exc:
                log.warning("file_sync: write failed for %s: %s", dst, exc)
    return written



"""
Multi-ring migration helpers shared between the runner subprocess
and the publish daemon.

Both surfaces need the same "drop unowned keys" logic:

* ``salt.runners.cluster.shed_unowned`` runs it from the operator's
  ``salt-run`` invocation.
* ``salt.channel.server.MasterPubServerChannel`` runs it on the
  daemon side when a peer's ``cluster/peer/shed-request`` event
  arrives (the fan-out path triggered by
  ``cluster.shed_unowned_all``).

Centralising the implementation here avoids the runner and the
daemon drifting apart on bank layout, cascade rules, or storage
replay quirks.  Both call sites pass their own ``__opts__`` dict in
explicitly so the helper has no loader dependency.
"""

import logging
import os
import time

log = logging.getLogger(__name__)


SHED_STATUS_FILENAME = "cluster-shed-status.json"


def perform_shed(
    opts,
    ring,
    banks=("jobs/loads", "jobs/minions", "jobs/endtimes", "jobs/nocache"),
    subbank_template="jobs/returns/{key}",
    driver=None,
    dry_run=False,
):
    """
    Drop the cache entries this master no longer owns under *ring*.

    Mirrors the operator-runner contract of
    ``salt.runners.cluster.shed_unowned`` but takes *opts* as an
    explicit argument so the publish daemon can call it without the
    loader's ``__opts__`` injection.

    Returns the same structured dict :func:`shed_unowned` returns
    (status / dropped / kept / subbanks_dropped / dry_run / ring).
    A "skipped" status carries a ``reason`` field; an "ok" status
    means the walk completed.
    """
    # Lazy imports — this module is loaded by the runner subprocess
    # which doesn't always have consensus deps available.
    import salt.cache  # pylint: disable=import-outside-toplevel
    from salt.cluster.consensus.raft.log import (  # pylint: disable=import-outside-toplevel
        RING_STATUS_ACTIVE,
        Log,
        LogEntryType,
        MembershipStateMachine,
        RingRegistryStateMachine,
    )
    from salt.cluster.consensus.storage import (  # pylint: disable=import-outside-toplevel
        SaltStorage,
    )
    from salt.cluster.ring import HashRing  # pylint: disable=import-outside-toplevel

    if not ring:
        raise ValueError("perform_shed requires a non-empty 'ring'")
    if not banks:
        raise ValueError("perform_shed requires at least one bank")

    node_id = opts.get("interface") or opts.get("id") or "unknown"

    # Cluster registry replay — same shape the runner uses.
    cluster_storage = SaltStorage(node_id, opts, ring_id="cluster")
    registry_sm = RingRegistryStateMachine()
    Log(
        storage=cluster_storage,
        state_machines={"ring_registry_sm": registry_sm},
    )
    for entry in cluster_storage.load_log():
        if entry.type == LogEntryType.RING_REGISTRY:
            registry_sm.apply(entry.cmd, index=entry.index)
    registry_entry = registry_sm.get(ring)
    if not registry_entry or registry_entry.get("status") != RING_STATUS_ACTIVE:
        return {
            "status": "skipped",
            "reason": f"ring {ring!r} is not active in the registry",
            "ring": ring,
            "dropped": 0,
            "kept": 0,
            "subbanks_dropped": 0,
            "dry_run": dry_run,
        }
    if node_id not in registry_entry.get("founding_voters", []):
        return {
            "status": "skipped",
            "reason": (
                f"this master ({node_id}) is not a founding voter of ring {ring!r}"
            ),
            "ring": ring,
            "dropped": 0,
            "kept": 0,
            "subbanks_dropped": 0,
            "dry_run": dry_run,
        }

    # Per-ring membership replay.
    ring_storage = SaltStorage(node_id, opts, ring_id=ring)
    ring_membership_sm = MembershipStateMachine()
    Log(
        storage=ring_storage,
        state_machines={"membership_sm": ring_membership_sm},
    )
    for entry in ring_storage.load_log():
        if entry.type == LogEntryType.CONFIG:
            ring_membership_sm.apply(entry.cmd, index=entry.index)
    voters = ring_membership_sm.current_voters() or registry_entry.get(
        "founding_voters", []
    )
    if not voters:
        return {
            "status": "skipped",
            "reason": f"ring {ring!r} has no committed voters yet",
            "ring": ring,
            "dropped": 0,
            "kept": 0,
            "subbanks_dropped": 0,
            "dry_run": dry_run,
        }

    hash_ring = HashRing()
    hash_ring.rebuild(voters)

    if driver is None:
        driver = opts.get("cache") or opts.get("keys.cache_driver")
    cache = salt.cache.Cache(opts, driver=driver)

    primary_bank = banks[0]
    unowned_primary_keys = []
    dropped, kept = 0, 0
    for idx, bank in enumerate(banks):
        try:
            keys = list(cache.list(bank))
        except Exception:  # pylint: disable=broad-except
            continue
        for key in keys:
            if hash_ring.owns(key, node_id):
                kept += 1
                continue
            if idx == 0:
                unowned_primary_keys.append(key)
            if not dry_run:
                try:
                    cache.flush(bank, key)
                except Exception:  # pylint: disable=broad-except
                    continue
            dropped += 1

    subbanks_dropped = 0
    if subbank_template and unowned_primary_keys:
        for key in unowned_primary_keys:
            subbank = subbank_template.format(key=key)
            if not dry_run:
                try:
                    cache.flush(subbank)
                except Exception:  # pylint: disable=broad-except
                    continue
            subbanks_dropped += 1

    log.info(
        "perform_shed: ring=%s dropped=%d kept=%d subbanks_dropped=%d "
        "dry_run=%s (primary_bank=%s)",
        ring,
        dropped,
        kept,
        subbanks_dropped,
        dry_run,
        primary_bank,
    )
    return {
        "status": "ok",
        "ring": ring,
        "dropped": dropped,
        "kept": kept,
        "subbanks_dropped": subbanks_dropped,
        "dry_run": dry_run,
    }


def write_shed_status(opts, result, source):
    """
    Persist a shed result for ``cluster.shed_status`` to surface.

    *source* explains who triggered the shed:

    * ``"runner"`` — this master ran ``cluster.shed_unowned``
      directly.
    * ``"runner_originator"`` — this master is the originator of a
      ``cluster.shed_unowned_all`` fan-out and ran its local pass.
    * ``"peer_request"`` — a peer's ``shed-request`` event reached
      this master.

    The sentinel is rewritten on every shed run.  Operators
    inspecting it always see the most-recent result.

    Writes are atomic — tmp file + rename — so a concurrent reader
    or a second writer mid-write never sees a partial JSON document.
    The bug this fixes: ``shed_unowned_all`` fan-out occasionally
    arrives twice at the same peer (see the ``self.pushers``
    duplication note in MULTI_RING_DESIGN.md), and the second write
    used to overwrite mid-stream and produce invalid JSON.
    """
    import json  # pylint: disable=import-outside-toplevel

    import salt.utils.atomicfile  # pylint: disable=import-outside-toplevel

    cachedir = opts.get("cachedir")
    if not cachedir:
        return
    path = os.path.join(cachedir, SHED_STATUS_FILENAME)
    body = dict(result)
    body["source"] = source
    body["updated_at"] = time.time()
    try:
        with salt.utils.atomicfile.atomic_open(path, "w") as fp:
            json.dump(body, fp)
    except OSError as exc:
        log.warning("shed-status: failed to write sentinel %s: %s", path, exc)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import ClassVar as _ClassVar

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pb2 as _descriptor_pb2
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor

class FieldBehavior(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    FIELD_BEHAVIOR_UNSPECIFIED: _ClassVar[FieldBehavior]
    OPTIONAL: _ClassVar[FieldBehavior]
    REQUIRED: _ClassVar[FieldBehavior]
    OUTPUT_ONLY: _ClassVar[FieldBehavior]
    INPUT_ONLY: _ClassVar[FieldBehavior]
    IMMUTABLE: _ClassVar[FieldBehavior]
    UNORDERED_LIST: _ClassVar[FieldBehavior]
    NON_EMPTY_DEFAULT: _ClassVar[FieldBehavior]
    IDENTIFIER: _ClassVar[FieldBehavior]

FIELD_BEHAVIOR_UNSPECIFIED: FieldBehavior
OPTIONAL: FieldBehavior
REQUIRED: FieldBehavior
OUTPUT_ONLY: FieldBehavior
INPUT_ONLY: FieldBehavior
IMMUTABLE: FieldBehavior
UNORDERED_LIST: FieldBehavior
NON_EMPTY_DEFAULT: FieldBehavior
IDENTIFIER: FieldBehavior
FIELD_BEHAVIOR_FIELD_NUMBER: _ClassVar[int]
field_behavior: _descriptor.FieldDescriptor



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2023, OpenTelemetry Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This file includes work covered by the following copyright and permission notices:

Copyright 2016 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import opentelemetry.proto.common.v1.common_pb2
import opentelemetry.proto.resource.v1.resource_pb2
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class ProfilesDictionary(google.protobuf.message.Message):
    """               Relationships Diagram

    ┌──────────────────┐                      LEGEND
    │   ProfilesData   │ ─────┐
    └──────────────────┘      │           ─────▶ embedded
      │                       │
      │ 1-n                   │           ─────▷ referenced by index
      ▼                       ▼
    ┌──────────────────┐   ┌────────────────────┐
    │ ResourceProfiles │   │ ProfilesDictionary │
    └──────────────────┘   └────────────────────┘
      │
      │ 1-n
      ▼
    ┌──────────────────┐
    │  ScopeProfiles   │
    └──────────────────┘
      │
      │ 1-n
      ▼
    ┌──────────────────┐
    │      Profile     │
    └──────────────────┘
      │                                n-1
      │ 1-n         ┌───────────────────────────────────────┐
      ▼             │                                       ▽
    ┌──────────────────┐   1-n   ┌─────────────────┐   ┌──────────┐
    │      Sample      │ ──────▷ │ KeyValueAndUnit │   │   Link   │
    └──────────────────┘         └─────────────────┘   └──────────┘
      │                              △      △
      │ n-1                          │      │ 1-n
      ▽                              │      │
    ┌──────────────────┐             │      │
    │      Stack       │             │      │
    └──────────────────┘             │      │
      │                     1-n      │      │
      │ 1-n         ┌────────────────┘      │
      ▽             │                       │
    ┌──────────────────┐   n-1   ┌─────────────┐
    │     Location     │ ──────▷ │   Mapping   │
    └──────────────────┘         └─────────────┘
      │
      │ 1-n
      ▼
    ┌──────────────────┐
    │       Line       │
    └──────────────────┘
      │
      │ 1-1
      ▽
    ┌──────────────────┐
    │     Function     │
    └──────────────────┘

    ProfilesDictionary represents the profiles data shared across the
    entire message being sent. The following applies to all fields in this
    message:

    - A dictionary is an array of dictionary items. Users of the dictionary
      compactly reference the items using the index within the array.

    - A dictionary MUST have a zero value encoded as the first element. This
      allows for _index fields pointing into the dictionary to use a 0 pointer
      value to indicate 'null' / 'not set'. Unless otherwise defined, a 'zero
      value' message value is one with all default field values, so as to
      minimize wire encoded size.

    - There SHOULD NOT be dupes in a dictionary. The identity of dictionary
      items is based on their value, recursively as needed. If a particular
      implementation does emit duplicated items, it MUST NOT attempt to give them
      meaning based on the index or order. A profile processor may remove
      duplicate items and this MUST NOT have any observable effects for
      consumers.

    - There SHOULD NOT be orphaned (unreferenced) items in a dictionary. A
      profile processor may remove ("garbage-collect") orphaned items and this
      MUST NOT have any observable effects for consumers.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAPPING_TABLE_FIELD_NUMBER: builtins.int
    LOCATION_TABLE_FIELD_NUMBER: builtins.int
    FUNCTION_TABLE_FIELD_NUMBER: builtins.int
    LINK_TABLE_FIELD_NUMBER: builtins.int
    STRING_TABLE_FIELD_NUMBER: builtins.int
    ATTRIBUTE_TABLE_FIELD_NUMBER: builtins.int
    STACK_TABLE_FIELD_NUMBER: builtins.int
    @property
    def mapping_table(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Mapping]:
        """Mappings from address ranges to the image/binary/library mapped
        into that address range referenced by locations via Location.mapping_index.

        mapping_table[0] must always be zero value (Mapping{}) and present.
        """
    @property
    def location_table(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Location]:
        """Locations referenced by samples via Stack.location_indices.

        location_table[0] must always be zero value (Location{}) and present.
        """
    @property
    def function_table(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Function]:
        """Functions referenced by locations via Line.function_index.

        function_table[0] must always be zero value (Function{}) and present.
        """
    @property
    def link_table(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Link]:
        """Links referenced by samples via Sample.link_index.

        link_table[0] must always be zero value (Link{}) and present.
        """
    @property
    def string_table(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """A common table for strings referenced by various messages.

        string_table[0] must always be "" and present.
        """
    @property
    def attribute_table(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyValueAndUnit]:
        """A common table for attributes referenced by the Profile, Sample, Mapping
        and Location messages below through attribute_indices field. Each entry is
        a key/value pair with an optional unit. Since this is a dictionary table,
        multiple entries with the same key may be present, unlike direct attribute
        tables like Resource.attributes. The referencing attribute_indices fields,
        though, do maintain the key uniqueness requirement.

        It's recommended to use attributes for variables with bounded cardinality,
        such as categorical variables
        (https://en.wikipedia.org/wiki/Categorical_variable). Using an attribute of
        a floating point type (e.g., CPU time) in a sample can quickly make every
        attribute value unique, defeating the purpose of the dictionary and
        impractically increasing the profile size.

        Examples of attributes:
            "/http/user_agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36"
            "abc.com/myattribute": true
            "allocation_size": 128 bytes

        attribute_table[0] must always be zero value (KeyValueAndUnit{}) and present.
        """
    @property
    def stack_table(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Stack]:
        """Stacks referenced by samples via Sample.stack_index.

        stack_table[0] must always be zero value (Stack{}) and present.
        """
    def __init__(
        self,
        *,
        mapping_table: collections.abc.Iterable[global___Mapping] | None = ...,
        location_table: collections.abc.Iterable[global___Location] | None = ...,
        function_table: collections.abc.Iterable[global___Function] | None = ...,
        link_table: collections.abc.Iterable[global___Link] | None = ...,
        string_table: collections.abc.Iterable[builtins.str] | None = ...,
        attribute_table: collections.abc.Iterable[global___KeyValueAndUnit] | None = ...,
        stack_table: collections.abc.Iterable[global___Stack] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_table", b"attribute_table", "function_table", b"function_table", "link_table", b"link_table", "location_table", b"location_table", "mapping_table", b"mapping_table", "stack_table", b"stack_table", "string_table", b"string_table"]) -> None: ...

global___ProfilesDictionary = ProfilesDictionary

@typing_extensions.final
class ProfilesData(google.protobuf.message.Message):
    """ProfilesData represents the profiles data that can be stored in persistent storage,
    OR can be embedded by other protocols that transfer OTLP profiles data but do not
    implement the OTLP protocol.

    The main difference between this message and collector protocol is that
    in this message there will not be any "control" or "metadata" specific to
    OTLP protocol.

    When new fields are added into this message, the OTLP request MUST be updated
    as well.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESOURCE_PROFILES_FIELD_NUMBER: builtins.int
    DICTIONARY_FIELD_NUMBER: builtins.int
    @property
    def resource_profiles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ResourceProfiles]:
        """An array of ResourceProfiles.
        For data coming from an SDK profiler, this array will typically contain one
        element. Host-level profilers will usually create one ResourceProfile per
        container, as well as one additional ResourceProfile grouping all samples
        from non-containerized processes.
        Other resource groupings are possible as well and clarified via
        Resource.attributes and semantic conventions.
        Tools that visualize profiles should prefer displaying
        resources_profiles[0].scope_profiles[0].profiles[0] by default.
        """
    @property
    def dictionary(self) -> global___ProfilesDictionary:
        """One instance of ProfilesDictionary"""
    def __init__(
        self,
        *,
        resource_profiles: collections.abc.Iterable[global___ResourceProfiles] | None = ...,
        dictionary: global___ProfilesDictionary | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["dictionary", b"dictionary"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["dictionary", b"dictionary", "resource_profiles", b"resource_profiles"]) -> None: ...

global___ProfilesData = ProfilesData

@typing_extensions.final
class ResourceProfiles(google.protobuf.message.Message):
    """A collection of ScopeProfiles from a Resource."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RESOURCE_FIELD_NUMBER: builtins.int
    SCOPE_PROFILES_FIELD_NUMBER: builtins.int
    SCHEMA_URL_FIELD_NUMBER: builtins.int
    @property
    def resource(self) -> opentelemetry.proto.resource.v1.resource_pb2.Resource:
        """The resource for the profiles in this message.
        If this field is not set then no resource info is known.
        """
    @property
    def scope_profiles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ScopeProfiles]:
        """A list of ScopeProfiles that originate from a resource."""
    schema_url: builtins.str
    """The Schema URL, if known. This is the identifier of the Schema that the resource data
    is recorded in. Notably, the last part of the URL path is the version number of the
    schema: http[s]://server[:port]/path/<version>. To learn more about Schema URL see
    https://opentelemetry.io/docs/specs/otel/schemas/#schema-url
    This schema_url applies to the data in the "resource" field. It does not apply
    to the data in the "scope_profiles" field which have their own schema_url field.
    """
    def __init__(
        self,
        *,
        resource: opentelemetry.proto.resource.v1.resource_pb2.Resource | None = ...,
        scope_profiles: collections.abc.Iterable[global___ScopeProfiles] | None = ...,
        schema_url: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["resource", b"resource"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["resource", b"resource", "schema_url", b"schema_url", "scope_profiles", b"scope_profiles"]) -> None: ...

global___ResourceProfiles = ResourceProfiles

@typing_extensions.final
class ScopeProfiles(google.protobuf.message.Message):
    """A collection of Profiles produced by an InstrumentationScope."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SCOPE_FIELD_NUMBER: builtins.int
    PROFILES_FIELD_NUMBER: builtins.int
    SCHEMA_URL_FIELD_NUMBER: builtins.int
    @property
    def scope(self) -> opentelemetry.proto.common.v1.common_pb2.InstrumentationScope:
        """The instrumentation scope information for the profiles in this message.
        Semantically when InstrumentationScope isn't set, it is equivalent with
        an empty instrumentation scope name (unknown).
        """
    @property
    def profiles(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Profile]:
        """A list of Profiles that originate from an instrumentation scope."""
    schema_url: builtins.str
    """The Schema URL, if known. This is the identifier of the Schema that the profile data
    is recorded in. Notably, the last part of the URL path is the version number of the
    schema: http[s]://server[:port]/path/<version>. To learn more about Schema URL see
    https://opentelemetry.io/docs/specs/otel/schemas/#schema-url
    This schema_url applies to the data in the "scope" field and all profiles in the
    "profiles" field.
    """
    def __init__(
        self,
        *,
        scope: opentelemetry.proto.common.v1.common_pb2.InstrumentationScope | None = ...,
        profiles: collections.abc.Iterable[global___Profile] | None = ...,
        schema_url: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["scope", b"scope"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["profiles", b"profiles", "schema_url", b"schema_url", "scope", b"scope"]) -> None: ...

global___ScopeProfiles = ScopeProfiles

@typing_extensions.final
class Profile(google.protobuf.message.Message):
    """Profile is a common stacktrace profile format.

    Measurements represented with this format should follow the
    following conventions:

    - Consumers should treat unset optional fields as if they had been
      set with their default value.

    - When possible, measurements should be stored in "unsampled" form
      that is most useful to humans.  There should be enough
      information present to determine the original sampled values.

    - The profile is represented as a set of samples, where each sample
      references a stack trace which is a list of locations, each belonging
      to a mapping.
    - There is a N->1 relationship from Stack.location_indices entries to
      locations. For every Stack.location_indices entry there must be a
      unique Location with that index.
    - There is an optional N->1 relationship from locations to
      mappings. For every nonzero Location.mapping_id there must be a
      unique Mapping with that index.

    Represents a complete profile, including sample types, samples, mappings to
    binaries, stacks, locations, functions, string table, and additional
    metadata. It modifies and annotates pprof Profile with OpenTelemetry
    specific fields.

    Note that whilst fields in this message retain the name and field id from pprof in most cases
    for ease of understanding data migration, it is not intended that pprof:Profile and
    OpenTelemetry:Profile encoding be wire compatible.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SAMPLE_TYPE_FIELD_NUMBER: builtins.int
    SAMPLES_FIELD_NUMBER: builtins.int
    TIME_UNIX_NANO_FIELD_NUMBER: builtins.int
    DURATION_NANO_FIELD_NUMBER: builtins.int
    PERIOD_TYPE_FIELD_NUMBER: builtins.int
    PERIOD_FIELD_NUMBER: builtins.int
    PROFILE_ID_FIELD_NUMBER: builtins.int
    DROPPED_ATTRIBUTES_COUNT_FIELD_NUMBER: builtins.int
    ORIGINAL_PAYLOAD_FORMAT_FIELD_NUMBER: builtins.int
    ORIGINAL_PAYLOAD_FIELD_NUMBER: builtins.int
    ATTRIBUTE_INDICES_FIELD_NUMBER: builtins.int
    @property
    def sample_type(self) -> global___ValueType:
        """The type and unit of all Sample.values in this profile.
        For a cpu or off-cpu profile this might be:
          ["cpu","nanoseconds"] or ["off_cpu","nanoseconds"]
        For a heap profile, this might be:
          ["allocated_objects","count"] or ["allocated_space","bytes"],
        """
    @property
    def samples(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Sample]:
        """The set of samples recorded in this profile."""
    time_unix_nano: builtins.int
    """The following fields 3-12 are informational, do not affect
    interpretation of results.

    Time of collection (UTC) represented as nanoseconds past the epoch.
    """
    duration_nano: builtins.int
    """Duration of the profile, if a duration makes sense."""
    @property
    def period_type(self) -> global___ValueType:
        """The kind of events between sampled occurrences.
        e.g [ "cpu","cycles" ] or [ "heap","bytes" ]
        """
    period: builtins.int
    """The number of events between sampled occurrences."""
    profile_id: builtins.bytes
    """A globally unique identifier for a profile. The ID is a 16-byte array. An ID with
    all zeroes is considered invalid. It may be used for deduplication and signal
    correlation purposes. It is acceptable to treat two profiles with different values
    in this field as not equal, even if they represented the same object at an earlier
    time.
    This field is optional; an ID may be assigned to an ID-less profile in a later step.
    """
    dropped_attributes_count: builtins.int
    """The number of attributes that were discarded. Attributes
    can be discarded because their keys are too long or because there are too many
    attributes. If this value is 0, then no attributes were dropped.
    """
    original_payload_format: builtins.str
    """The original payload format. See also original_payload. Optional, but the
    format and the bytes must be set or unset together.

    The allowed values for the format string are defined by the OpenTelemetry
    specification. Some examples are "jfr", "pprof", "linux_perf".

    The original payload may be optionally provided when the conversion to the
    OLTP format was done from a different format with some loss of the fidelity
    and the receiver may want to store the original payload to allow future
    lossless export or reinterpretation. Some examples of the original format
    are JFR (Java Flight Recorder), pprof, Linux perf.

    Even when the original payload is in a format that is semantically close to
    OTLP, such as pprof, a conversion may still be lossy in some cases (e.g. if
    the pprof file contains custom extensions or conventions).

    The original payload can be large in size, so including the original
    payload should be configurable by the profiler or collector options. The
    default behavior should be to not include the original payload.
    """
    original_payload: builtins.bytes
    """The original payload bytes. See also original_payload_format. Optional, but
    format and the bytes must be set or unset together.
    """
    @property
    def attribute_indices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """References to attributes in attribute_table. [optional]"""
    def __init__(
        self,
        *,
        sample_type: global___ValueType | None = ...,
        samples: collections.abc.Iterable[global___Sample] | None = ...,
        time_unix_nano: builtins.int = ...,
        duration_nano: builtins.int = ...,
        period_type: global___ValueType | None = ...,
        period: builtins.int = ...,
        profile_id: builtins.bytes = ...,
        dropped_attributes_count: builtins.int = ...,
        original_payload_format: builtins.str = ...,
        original_payload: builtins.bytes = ...,
        attribute_indices: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["period_type", b"period_type", "sample_type", b"sample_type"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_indices", b"attribute_indices", "dropped_attributes_count", b"dropped_attributes_count", "duration_nano", b"duration_nano", "original_payload", b"original_payload", "original_payload_format", b"original_payload_format", "period", b"period", "period_type", b"period_type", "profile_id", b"profile_id", "sample_type", b"sample_type", "samples", b"samples", "time_unix_nano", b"time_unix_nano"]) -> None: ...

global___Profile = Profile

@typing_extensions.final
class Link(google.protobuf.message.Message):
    """A pointer from a profile Sample to a trace Span.
    Connects a profile sample to a trace span, identified by unique trace and span IDs.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TRACE_ID_FIELD_NUMBER: builtins.int
    SPAN_ID_FIELD_NUMBER: builtins.int
    trace_id: builtins.bytes
    """A unique identifier of a trace that this linked span is part of. The ID is a
    16-byte array.
    """
    span_id: builtins.bytes
    """A unique identifier for the linked span. The ID is an 8-byte array."""
    def __init__(
        self,
        *,
        trace_id: builtins.bytes = ...,
        span_id: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["span_id", b"span_id", "trace_id", b"trace_id"]) -> None: ...

global___Link = Link

@typing_extensions.final
class ValueType(google.protobuf.message.Message):
    """ValueType describes the type and units of a value."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TYPE_STRINDEX_FIELD_NUMBER: builtins.int
    UNIT_STRINDEX_FIELD_NUMBER: builtins.int
    type_strindex: builtins.int
    """Index into ProfilesDictionary.string_table."""
    unit_strindex: builtins.int
    """Index into ProfilesDictionary.string_table."""
    def __init__(
        self,
        *,
        type_strindex: builtins.int = ...,
        unit_strindex: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["type_strindex", b"type_strindex", "unit_strindex", b"unit_strindex"]) -> None: ...

global___ValueType = ValueType

@typing_extensions.final
class Sample(google.protobuf.message.Message):
    """Each Sample records values encountered in some program context. The program
    context is typically a stack trace, perhaps augmented with auxiliary
    information like the thread-id, some indicator of a higher level request
    being handled etc.

    A Sample MUST have have at least one values or timestamps_unix_nano entry. If
    both fields are populated, they MUST contain the same number of elements, and
    the elements at the same index MUST refer to the same event.

    Examples of different ways of representing a sample with the total value of 10:

    Report of a stacktrace at 10 timestamps (consumers must assume the value is 1 for each point):
       values: []
       timestamps_unix_nano: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]

    Report of a stacktrace with an aggregated value without timestamps:
      values: [10]
       timestamps_unix_nano: []

    Report of a stacktrace at 4 timestamps where each point records a specific value:
       values: [2, 2, 3, 3]
       timestamps_unix_nano: [1, 2, 3, 4]
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STACK_INDEX_FIELD_NUMBER: builtins.int
    VALUES_FIELD_NUMBER: builtins.int
    ATTRIBUTE_INDICES_FIELD_NUMBER: builtins.int
    LINK_INDEX_FIELD_NUMBER: builtins.int
    TIMESTAMPS_UNIX_NANO_FIELD_NUMBER: builtins.int
    stack_index: builtins.int
    """Reference to stack in ProfilesDictionary.stack_table."""
    @property
    def values(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """The type and unit of each value is defined by Profile.sample_type."""
    @property
    def attribute_indices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """References to attributes in ProfilesDictionary.attribute_table. [optional]"""
    link_index: builtins.int
    """Reference to link in ProfilesDictionary.link_table. [optional]
    It can be unset / set to 0 if no link exists, as link_table[0] is always a 'null' default value.
    """
    @property
    def timestamps_unix_nano(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Timestamps associated with Sample represented in nanoseconds. These
        timestamps should fall within the Profile's time range.
        """
    def __init__(
        self,
        *,
        stack_index: builtins.int = ...,
        values: collections.abc.Iterable[builtins.int] | None = ...,
        attribute_indices: collections.abc.Iterable[builtins.int] | None = ...,
        link_index: builtins.int = ...,
        timestamps_unix_nano: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_indices", b"attribute_indices", "link_index", b"link_index", "stack_index", b"stack_index", "timestamps_unix_nano", b"timestamps_unix_nano", "values", b"values"]) -> None: ...

global___Sample = Sample

@typing_extensions.final
class Mapping(google.protobuf.message.Message):
    """Describes the mapping of a binary in memory, including its address range,
    file offset, and metadata like build ID
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MEMORY_START_FIELD_NUMBER: builtins.int
    MEMORY_LIMIT_FIELD_NUMBER: builtins.int
    FILE_OFFSET_FIELD_NUMBER: builtins.int
    FILENAME_STRINDEX_FIELD_NUMBER: builtins.int
    ATTRIBUTE_INDICES_FIELD_NUMBER: builtins.int
    memory_start: builtins.int
    """Address at which the binary (or DLL) is loaded into memory."""
    memory_limit: builtins.int
    """The limit of the address range occupied by this mapping."""
    file_offset: builtins.int
    """Offset in the binary that corresponds to the first mapped address."""
    filename_strindex: builtins.int
    """The object this entry is loaded from.  This can be a filename on
    disk for the main binary and shared libraries, or virtual
    abstractions like "[vdso]".
    Index into ProfilesDictionary.string_table.
    """
    @property
    def attribute_indices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """References to attributes in ProfilesDictionary.attribute_table. [optional]"""
    def __init__(
        self,
        *,
        memory_start: builtins.int = ...,
        memory_limit: builtins.int = ...,
        file_offset: builtins.int = ...,
        filename_strindex: builtins.int = ...,
        attribute_indices: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["attribute_indices", b"attribute_indices", "file_offset", b"file_offset", "filename_strindex", b"filename_strindex", "memory_limit", b"memory_limit", "memory_start", b"memory_start"]) -> None: ...

global___Mapping = Mapping

@typing_extensions.final
class Stack(google.protobuf.message.Message):
    """A Stack represents a stack trace as a list of locations."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LOCATION_INDICES_FIELD_NUMBER: builtins.int
    @property
    def location_indices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """References to locations in ProfilesDictionary.location_table.
        The first location is the leaf frame.
        """
    def __init__(
        self,
        *,
        location_indices: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["location_indices", b"location_indices"]) -> None: ...

global___Stack = Stack

@typing_extensions.final
class Location(google.protobuf.message.Message):
    """Describes function and line table debug information."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MAPPING_INDEX_FIELD_NUMBER: builtins.int
    ADDRESS_FIELD_NUMBER: builtins.int
    LINES_FIELD_NUMBER: builtins.int
    ATTRIBUTE_INDICES_FIELD_NUMBER: builtins.int
    mapping_index: builtins.int
    """Reference to mapping in ProfilesDictionary.mapping_table.
    It can be unset / set to 0 if the mapping is unknown or not applicable for
    this profile type, as mapping_table[0] is always a 'null' default mapping.
    """
    address: builtins.int
    """The instruction address for this location, if available.  It
    should be within [Mapping.memory_start...Mapping.memory_limit]
    for the corresponding mapping. A non-leaf address may be in the
    middle of a call instruction. It is up to display tools to find
    the beginning of the instruction if necessary.
    """
    @property
    def lines(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Line]:
        """Multiple line indicates this location has inlined functions,
        where the last entry represents the caller into which the
        preceding entries were inlined.

        E.g., if memcpy() is inlined into printf:
           lines[0].function_name == "memcpy"
           lines[1].function_name == "printf"
        """
    @property
    def attribute_indices(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """References to attributes in ProfilesDictionary.attribute_table. [optional]"""
    def __init__(
        self,
        *,
        mapping_index: builtins.int = ...,
        address: builtins.int = ...,
        lines: collections.abc.Iterable[global___Line] | None = ...,
        attribute_indices: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["address", b"address", "attribute_indices", b"attribute_indices", "lines", b"lines", "mapping_index", b"mapping_index"]) -> None: ...

global___Location = Location

@typing_extensions.final
class Line(google.protobuf.message.Message):
    """Details a specific line in a source code, linked to a function."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FUNCTION_INDEX_FIELD_NUMBER: builtins.int
    LINE_FIELD_NUMBER: builtins.int
    COLUMN_FIELD_NUMBER: builtins.int
    function_index: builtins.int
    """Reference to function in ProfilesDictionary.function_table."""
    line: builtins.int
    """Line number in source code. 0 means unset."""
    column: builtins.int
    """Column number in source code. 0 means unset."""
    def __init__(
        self,
        *,
        function_index: builtins.int = ...,
        line: builtins.int = ...,
        column: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["column", b"column", "function_index", b"function_index", "line", b"line"]) -> None: ...

global___Line = Line

@typing_extensions.final
class Function(google.protobuf.message.Message):
    """Describes a function, including its human-readable name, system name,
    source file, and starting line number in the source.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_STRINDEX_FIELD_NUMBER: builtins.int
    SYSTEM_NAME_STRINDEX_FIELD_NUMBER: builtins.int
    FILENAME_STRINDEX_FIELD_NUMBER: builtins.int
    START_LINE_FIELD_NUMBER: builtins.int
    name_strindex: builtins.int
    """The function name. Empty string if not available."""
    system_name_strindex: builtins.int
    """Function name, as identified by the system. For instance,
    it can be a C++ mangled name. Empty string if not available.
    """
    filename_strindex: builtins.int
    """Source file containing the function. Empty string if not available."""
    start_line: builtins.int
    """Line number in source file. 0 means unset."""
    def __init__(
        self,
        *,
        name_strindex: builtins.int = ...,
        system_name_strindex: builtins.int = ...,
        filename_strindex: builtins.int = ...,
        start_line: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["filename_strindex", b"filename_strindex", "name_strindex", b"name_strindex", "start_line", b"start_line", "system_name_strindex", b"system_name_strindex"]) -> None: ...

global___Function = Function

@typing_extensions.final
class KeyValueAndUnit(google.protobuf.message.Message):
    """A custom 'dictionary native' style of encoding attributes which is more convenient
    for profiles than opentelemetry.proto.common.v1.KeyValue
    Specifically, uses the string table for keys and allows optional unit information.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    KEY_STRINDEX_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    UNIT_STRINDEX_FIELD_NUMBER: builtins.int
    key_strindex: builtins.int
    """The index into the string table for the attribute's key."""
    @property
    def value(self) -> opentelemetry.proto.common.v1.common_pb2.AnyValue:
        """The value of the attribute."""
    unit_strindex: builtins.int
    """The index into the string table for the attribute's unit.
    zero indicates implicit (by semconv) or non-defined unit.
    """
    def __init__(
        self,
        *,
        key_strindex: builtins.int = ...,
        value: opentelemetry.proto.common.v1.common_pb2.AnyValue | None = ...,
        unit_strindex: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["key_strindex", b"key_strindex", "unit_strindex", b"unit_strindex", "value", b"value"]) -> None: ...

global___KeyValueAndUnit = KeyValueAndUnit



"""Disassembler of Python byte code into mnemonics."""

import sys
import types
import collections
import io

from opcode import *
from opcode import (
    __all__ as _opcodes_all,
    _cache_format,
    _inline_cache_entries,
    _nb_ops,
    _common_constants,
    _intrinsic_1_descs,
    _intrinsic_2_descs,
    _special_method_names,
    _specializations,
    _specialized_opmap,
)

from _opcode import get_executor

__all__ = ["code_info", "dis", "disassemble", "distb", "disco",
           "findlinestarts", "findlabels", "show_code",
           "get_instructions", "Instruction", "Bytecode"] + _opcodes_all
del _opcodes_all

_have_code = (types.MethodType, types.FunctionType, types.CodeType,
              classmethod, staticmethod, type)

CONVERT_VALUE = opmap['CONVERT_VALUE']

SET_FUNCTION_ATTRIBUTE = opmap['SET_FUNCTION_ATTRIBUTE']
FUNCTION_ATTR_FLAGS = ('defaults', 'kwdefaults', 'annotations', 'closure', 'annotate')

ENTER_EXECUTOR = opmap['ENTER_EXECUTOR']
LOAD_GLOBAL = opmap['LOAD_GLOBAL']
LOAD_SMALL_INT = opmap['LOAD_SMALL_INT']
BINARY_OP = opmap['BINARY_OP']
JUMP_BACKWARD = opmap['JUMP_BACKWARD']
FOR_ITER = opmap['FOR_ITER']
SEND = opmap['SEND']
LOAD_ATTR = opmap['LOAD_ATTR']
LOAD_SUPER_ATTR = opmap['LOAD_SUPER_ATTR']
CALL_INTRINSIC_1 = opmap['CALL_INTRINSIC_1']
CALL_INTRINSIC_2 = opmap['CALL_INTRINSIC_2']
LOAD_COMMON_CONSTANT = opmap['LOAD_COMMON_CONSTANT']
LOAD_SPECIAL = opmap['LOAD_SPECIAL']
LOAD_FAST_LOAD_FAST = opmap['LOAD_FAST_LOAD_FAST']
LOAD_FAST_BORROW_LOAD_FAST_BORROW = opmap['LOAD_FAST_BORROW_LOAD_FAST_BORROW']
STORE_FAST_LOAD_FAST = opmap['STORE_FAST_LOAD_FAST']
STORE_FAST_STORE_FAST = opmap['STORE_FAST_STORE_FAST']
IS_OP = opmap['IS_OP']
CONTAINS_OP = opmap['CONTAINS_OP']
END_ASYNC_FOR = opmap['END_ASYNC_FOR']

CACHE = opmap["CACHE"]

_all_opname = list(opname)
_all_opmap = dict(opmap)
for name, op in _specialized_opmap.items():
    # fill opname and opmap
    assert op < len(_all_opname)
    _all_opname[op] = name
    _all_opmap[name] = op

deoptmap = {
    specialized: base for base, family in _specializations.items() for specialized in family
}

def _try_compile(source, name):
    """Attempts to compile the given source, first as an expression and
       then as a statement if the first approach fails.

       Utility function to accept strings in functions that otherwise
       expect code objects
    """
    try:
        return compile(source, name, 'eval')
    except SyntaxError:
        pass
    return compile(source, name, 'exec')

def dis(x=None, *, file=None, depth=None, show_caches=False, adaptive=False,
        show_offsets=False, show_positions=False):
    """Disassemble classes, methods, functions, and other compiled objects.

    With no argument, disassemble the last traceback.

    Compiled objects currently include generator objects, async generator
    objects, and coroutine objects, all of which store their code object
    in a special attribute.
    """
    if x is None:
        distb(file=file, show_caches=show_caches, adaptive=adaptive,
              show_offsets=show_offsets, show_positions=show_positions)
        return
    # Extract functions from methods.
    if hasattr(x, '__func__'):
        x = x.__func__
    # Extract compiled code objects from...
    if hasattr(x, '__code__'):  # ...a function, or
        x = x.__code__
    elif hasattr(x, 'gi_code'):  #...a generator object, or
        x = x.gi_code
    elif hasattr(x, 'ag_code'):  #...an asynchronous generator object, or
        x = x.ag_code
    elif hasattr(x, 'cr_code'):  #...a coroutine.
        x = x.cr_code
    # Perform the disassembly.
    if hasattr(x, '__dict__'):  # Class or module
        items = sorted(x.__dict__.items())
        for name, x1 in items:
            if isinstance(x1, _have_code):
                print("Disassembly of %s:" % name, file=file)
                try:
                    dis(x1, file=file, depth=depth, show_caches=show_caches, adaptive=adaptive, show_offsets=show_offsets, show_positions=show_positions)
                except TypeError as msg:
                    print("Sorry:", msg, file=file)
                print(file=file)
    elif hasattr(x, 'co_code'): # Code object
        _disassemble_recursive(x, file=file, depth=depth, show_caches=show_caches, adaptive=adaptive, show_offsets=show_offsets, show_positions=show_positions)
    elif isinstance(x, (bytes, bytearray)): # Raw bytecode
        labels_map = _make_labels_map(x)
        label_width = 4 + len(str(len(labels_map)))
        formatter = Formatter(file=file,
                              offset_width=len(str(max(len(x) - 2, 9999))) if show_offsets else 0,
                              label_width=label_width,
                              show_caches=show_caches)
        arg_resolver = ArgResolver(labels_map=labels_map)
        _disassemble_bytes(x, arg_resolver=arg_resolver, formatter=formatter)
    elif isinstance(x, str):    # Source code
        _disassemble_str(x, file=file, depth=depth, show_caches=show_caches, adaptive=adaptive, show_offsets=show_offsets, show_positions=show_positions)
    else:
        raise TypeError("don't know how to disassemble %s objects" %
                        type(x).__name__)

def distb(tb=None, *, file=None, show_caches=False, adaptive=False, show_offsets=False, show_positions=False):
    """Disassemble a traceback (default: last traceback)."""
    if tb is None:
        try:
            if hasattr(sys, 'last_exc'):
                tb = sys.last_exc.__traceback__
            else:
                tb = sys.last_traceback
        except AttributeError:
            raise RuntimeError("no last traceback to disassemble") from None
        while tb.tb_next: tb = tb.tb_next
    disassemble(tb.tb_frame.f_code, tb.tb_lasti, file=file, show_caches=show_caches, adaptive=adaptive, show_offsets=show_offsets, show_positions=show_positions)

# The inspect module interrogates this dictionary to build its
# list of CO_* constants. It is also used by pretty_flags to
# turn the co_flags field into a human readable list.
COMPILER_FLAG_NAMES = {
            1: "OPTIMIZED",
            2: "NEWLOCALS",
            4: "VARARGS",
            8: "VARKEYWORDS",
           16: "NESTED",
           32: "GENERATOR",
           64: "NOFREE",
          128: "COROUTINE",
          256: "ITERABLE_COROUTINE",
          512: "ASYNC_GENERATOR",
    0x4000000: "HAS_DOCSTRING",
    0x8000000: "METHOD",
}

def pretty_flags(flags):
    """Return pretty representation of code flags."""
    names = []
    for i in range(32):
        flag = 1<<i
        if flags & flag:
            names.append(COMPILER_FLAG_NAMES.get(flag, hex(flag)))
            flags ^= flag
            if not flags:
                break
    else:
        names.append(hex(flags))
    return ", ".join(names)

class _Unknown:
    def __repr__(self):
        return "<unknown>"

# Sentinel to represent values that cannot be calculated
UNKNOWN = _Unknown()

def _get_code_object(x):
    """Helper to handle methods, compiled or raw code objects, and strings."""
    # Extract functions from methods.
    if hasattr(x, '__func__'):
        x = x.__func__
    # Extract compiled code objects from...
    if hasattr(x, '__code__'):  # ...a function, or
        x = x.__code__
    elif hasattr(x, 'gi_code'):  #...a generator object, or
        x = x.gi_code
    elif hasattr(x, 'ag_code'):  #...an asynchronous generator object, or
        x = x.ag_code
    elif hasattr(x, 'cr_code'):  #...a coroutine.
        x = x.cr_code
    # Handle source code.
    if isinstance(x, str):
        x = _try_compile(x, "<disassembly>")
    # By now, if we don't have a code object, we can't disassemble x.
    if hasattr(x, 'co_code'):
        return x
    raise TypeError("don't know how to disassemble %s objects" %
                    type(x).__name__)

def _deoptop(op):
    name = _all_opname[op]
    return _all_opmap[deoptmap[name]] if name in deoptmap else op

def _get_code_array(co, adaptive):
    if adaptive:
        code = co._co_code_adaptive
        res = []
        found = False
        for i in range(0, len(code), 2):
            op, arg = code[i], code[i+1]
            if op == ENTER_EXECUTOR:
                try:
                    ex = get_executor(co, i)
                except (ValueError, RuntimeError):
                    ex = None

                if ex:
                    op, arg = ex.get_opcode(), ex.get_oparg()
                    found = True

            res.append(op.to_bytes())
            res.append(arg.to_bytes())
        return code if not found else b''.join(res)
    else:
        return co.co_code

def code_info(x):
    """Formatted details of methods, functions, or code."""
    return _format_code_info(_get_code_object(x))

def _format_code_info(co):
    lines = []
    lines.append("Name:              %s" % co.co_name)
    lines.append("Filename:          %s" % co.co_filename)
    lines.append("Argument count:    %s" % co.co_argcount)
    lines.append("Positional-only arguments: %s" % co.co_posonlyargcount)
    lines.append("Kw-only arguments: %s" % co.co_kwonlyargcount)
    lines.append("Number of locals:  %s" % co.co_nlocals)
    lines.append("Stack size:        %s" % co.co_stacksize)
    lines.append("Flags:             %s" % pretty_flags(co.co_flags))
    if co.co_consts:
        lines.append("Constants:")
        for i_c in enumerate(co.co_consts):
            lines.append("%4d: %r" % i_c)
    if co.co_names:
        lines.append("Names:")
        for i_n in enumerate(co.co_names):
            lines.append("%4d: %s" % i_n)
    if co.co_varnames:
        lines.append("Variable names:")
        for i_n in enumerate(co.co_varnames):
            lines.append("%4d: %s" % i_n)
    if co.co_freevars:
        lines.append("Free variables:")
        for i_n in enumerate(co.co_freevars):
            lines.append("%4d: %s" % i_n)
    if co.co_cellvars:
        lines.append("Cell variables:")
        for i_n in enumerate(co.co_cellvars):
            lines.append("%4d: %s" % i_n)
    return "\n".join(lines)

def show_code(co, *, file=None):
    """Print details of methods, functions, or code to *file*.

    If *file* is not provided, the output is printed on stdout.
    """
    print(code_info(co), file=file)

Positions = collections.namedtuple(
    'Positions',
    [
        'lineno',
        'end_lineno',
        'col_offset',
        'end_col_offset',
    ],
    defaults=[None] * 4
)

_Instruction = collections.namedtuple(
    "_Instruction",
    [
        'opname',
        'opcode',
        'arg',
        'argval',
        'argrepr',
        'offset',
        'start_offset',
        'starts_line',
        'line_number',
        'label',
        'positions',
        'cache_info',
    ],
    defaults=[None, None, None]
)

_Instruction.opname.__doc__ = "Human readable name for operation"
_Instruction.opcode.__doc__ = "Numeric code for operation"
_Instruction.arg.__doc__ = "Numeric argument to operation (if any), otherwise None"
_Instruction.argval.__doc__ = "Resolved arg value (if known), otherwise same as arg"
_Instruction.argrepr.__doc__ = "Human readable description of operation argument"
_Instruction.offset.__doc__ = "Start index of operation within bytecode sequence"
_Instruction.start_offset.__doc__ = (
    "Start index of operation within bytecode sequence, including extended args if present; "
    "otherwise equal to Instruction.offset"
)
_Instruction.starts_line.__doc__ = "True if this opcode starts a source line, otherwise False"
_Instruction.line_number.__doc__ = "source line number associated with this opcode (if any), otherwise None"
_Instruction.label.__doc__ = "A label (int > 0) if this instruction is a jump target, otherwise None"
_Instruction.positions.__doc__ = "dis.Positions object holding the span of source code covered by this instruction"
_Instruction.cache_info.__doc__ = "list of (name, size, data), one for each cache entry of the instruction"

_ExceptionTableEntryBase = collections.namedtuple("_ExceptionTableEntryBase",
    "start end target depth lasti")

class _ExceptionTableEntry(_ExceptionTableEntryBase):
    pass

_OPNAME_WIDTH = 20
_OPARG_WIDTH = 5

def _get_cache_size(opname):
    return _inline_cache_entries.get(opname, 0)

def _get_jump_target(op, arg, offset):
    """Gets the bytecode offset of the jump target if this is a jump instruction.

    Otherwise return None.
    """
    deop = _deoptop(op)
    caches = _get_cache_size(_all_opname[deop])
    if deop in hasjrel:
        if _is_backward_jump(deop):
            arg = -arg
        target = offset + 2 + arg*2
        target += 2 * caches
    elif deop in hasjabs:
        target = arg*2
    else:
        target = None
    return target

class Instruction(_Instruction):
    """Details for a bytecode operation.

       Defined fields:
         opname - human readable name for operation
         opcode - numeric code for operation
         arg - numeric argument to operation (if any), otherwise None
         argval - resolved arg value (if known), otherwise same as arg
         argrepr - human readable description of operation argument
         offset - start index of operation within bytecode sequence
         start_offset - start index of operation within bytecode sequence including extended args if present;
                        otherwise equal to Instruction.offset
         starts_line - True if this opcode starts a source line, otherwise False
         line_number - source line number associated with this opcode (if any), otherwise None
         label - A label if this instruction is a jump target, otherwise None
         positions - Optional dis.Positions object holding the span of source code
                     covered by this instruction
         cache_info - information about the format and content of the instruction's cache
                        entries (if any)
    """

    @staticmethod
    def make(
        opname, arg, argval, argrepr, offset, start_offset, starts_line,
        line_number, label=None, positions=None, cache_info=None
    ):
        return Instruction(opname, _all_opmap[opname], arg, argval, argrepr, offset,
                           start_offset, starts_line, line_number, label, positions, cache_info)

    @property
    def oparg(self):
        """Alias for Instruction.arg."""
        return self.arg

    @property
    def baseopcode(self):
        """Numeric code for the base operation if operation is specialized.

        Otherwise equal to Instruction.opcode.
        """
        return _deoptop(self.opcode)

    @property
    def baseopname(self):
        """Human readable name for the base operation if operation is specialized.

        Otherwise equal to Instruction.opname.
        """
        return opname[self.baseopcode]

    @property
    def cache_offset(self):
        """Start index of the cache entries following the operation."""
        return self.offset + 2

    @property
    def end_offset(self):
        """End index of the cache entries following the operation."""
        return self.cache_offset + _get_cache_size(_all_opname[self.opcode])*2

    @property
    def jump_target(self):
        """Bytecode index of the jump target if this is a jump operation.

        Otherwise return None.
        """
        return _get_jump_target(self.opcode, self.arg, self.offset)

    @property
    def is_jump_target(self):
        """True if other code jumps to here, otherwise False"""
        return self.label is not None

    def __str__(self):
        output = io.StringIO()
        formatter = Formatter(file=output)
        formatter.print_instruction(self, False)
        return output.getvalue()


class Formatter:

    def __init__(self, file=None, lineno_width=0, offset_width=0, label_width=0,
                 line_offset=0, show_caches=False, *, show_positions=False):
        """Create a Formatter

        *file* where to write the output
        *lineno_width* sets the width of the source location field (0 omits it).
        Should be large enough for a line number or full positions (depending
        on the value of *show_positions*).
        *offset_width* sets the width of the instruction offset field
        *label_width* sets the width of the label field
        *show_caches* is a boolean indicating whether to display cache lines
        *show_positions* is a boolean indicating whether full positions should
        be reported instead of only the line numbers.
        """
        self.file = file
        self.lineno_width = lineno_width
        self.offset_width = offset_width
        self.label_width = label_width
        self.show_caches = show_caches
        self.show_positions = show_positions

    def print_instruction(self, instr, mark_as_current=False):
        self.print_instruction_line(instr, mark_as_current)
        if self.show_caches and instr.cache_info:
            offset = instr.offset
            for name, size, data in instr.cache_info:
                for i in range(size):
                    offset += 2
                    # Only show the fancy argrepr for a CACHE instruction when it's
                    # the first entry for a particular cache value:
                    if i == 0:
                        argrepr = f"{name}: {int.from_bytes(data, sys.byteorder)}"
                    else:
                        argrepr = ""
                    self.print_instruction_line(
                        Instruction("CACHE", CACHE, 0, None, argrepr, offset, offset,
                                    False, None, None, instr.positions),
                        False)

    def print_instruction_line(self, instr, mark_as_current):
        """Format instruction details for inclusion in disassembly output."""
        lineno_width = self.lineno_width
        offset_width = self.offset_width
        label_width = self.label_width

        new_source_line = (lineno_width > 0 and
                           instr.starts_line and
                           instr.offset > 0)
        if new_source_line:
            print(file=self.file)

        fields = []
        # Column: Source code locations information
        if lineno_width:
            if self.show_positions:
                # reporting positions instead of just line numbers
                if instr_positions := instr.positions:
                    if all(p is None for p in instr_positions):
                        positions_str = _NO_LINENO
                    else:
                        ps = tuple('?' if p is None else p for p in instr_positions)
                        positions_str = f"{ps[0]}:{ps[2]}-{ps[1]}:{ps[3]}"
                    fields.append(f'{positions_str:{lineno_width}}')
                else:
                    fields.append(' ' * lineno_width)
            else:
                if instr.starts_line:
                    lineno_fmt = "%%%dd" if instr.line_number is not None else "%%%ds"
                    lineno_fmt = lineno_fmt % lineno_width
                    lineno = _NO_LINENO if instr.line_number is None else instr.line_number
                    fields.append(lineno_fmt % lineno)
                else:
                    fields.append(' ' * lineno_width)
        # Column: Label
        if instr.label is not None:
            lbl = f"L{instr.label}:"
            fields.append(f"{lbl:>{label_width}}")
        else:
            fields.append(' ' * label_width)
        # Column: Instruction offset from start of code sequence
        if offset_width > 0:
            fields.append(f"{repr(instr.offset):>{offset_width}}  ")
        # Column: Current instruction indicator
        if mark_as_current:
            fields.append('-->')
        else:
            fields.append('   ')
        # Column: Opcode name
        fields.append(instr.opname.ljust(_OPNAME_WIDTH))
        # Column: Opcode argument
        if instr.arg is not None:
            arg = repr(instr.arg)
            # If opname is longer than _OPNAME_WIDTH, we allow it to overflow into
            # the space reserved for oparg. This results in fewer misaligned opargs
            # in the disassembly output.
            opname_excess = max(0, len(instr.opname) - _OPNAME_WIDTH)
            fields.append(repr(instr.arg).rjust(_OPARG_WIDTH - opname_excess))
            # Column: Opcode argument details
            if instr.argrepr:
                fields.append('(' + instr.argrepr + ')')
        print(' '.join(fields).rstrip(), file=self.file)

    def print_exception_table(self, exception_entries):
        file = self.file
        if exception_entries:
            print("ExceptionTable:", file=file)
            for entry in exception_entries:
                lasti = " lasti" if entry.lasti else ""
                start = entry.start_label
                end = entry.end_label
                target = entry.target_label
                print(f"  L{start} to L{end} -> L{target} [{entry.depth}]{lasti}", file=file)


class ArgResolver:
    def __init__(self, co_consts=None, names=None, varname_from_oparg=None, labels_map=None):
        self.co_consts = co_consts
        self.names = names
        self.varname_from_oparg = varname_from_oparg
        self.labels_map = labels_map or {}

    def offset_from_jump_arg(self, op, arg, offset):
        deop = _deoptop(op)
        if deop in hasjabs:
            return arg * 2
        elif deop in hasjrel:
            signed_arg = -arg if _is_backward_jump(deop) else arg
            argval = offset + 2 + signed_arg*2
            caches = _get_cache_size(_all_opname[deop])
            argval += 2 * caches
            return argval
        return None

    def get_label_for_offset(self, offset):
        return self.labels_map.get(offset, None)

    def get_argval_argrepr(self, op, arg, offset):
        get_name = None if self.names is None else self.names.__getitem__
        argval = None
        argrepr = ''
        deop = _deoptop(op)
        if arg is not None:
            #  Set argval to the dereferenced value of the argument when
            #  available, and argrepr to the string representation of argval.
            #    _disassemble_bytes needs the string repr of the
            #    raw name index for LOAD_GLOBAL, LOAD_CONST, etc.
            argval = arg
            if deop in hasconst:
                argval, argrepr = _get_const_info(deop, arg, self.co_consts)
            elif deop in hasname:
                if deop == LOAD_GLOBAL:
                    argval, argrepr = _get_name_info(arg//2, get_name)
                    if (arg & 1) and argrepr:
                        argrepr = f"{argrepr} + NULL"
                elif deop == LOAD_ATTR:
                    argval, argrepr = _get_name_info(arg//2, get_name)
                    if (arg & 1) and argrepr:
                        argrepr = f"{argrepr} + NULL|self"
                elif deop == LOAD_SUPER_ATTR:
                    argval, argrepr = _get_name_info(arg//4, get_name)
                    if (arg & 1) and argrepr:
                        argrepr = f"{argrepr} + NULL|self"
                else:
                    argval, argrepr = _get_name_info(arg, get_name)
            elif deop in hasjump or deop in hasexc:
                argval = self.offset_from_jump_arg(op, arg, offset)
                lbl = self.get_label_for_offset(argval)
                assert lbl is not None
                preposition = "from" if deop == END_ASYNC_FOR else "to"
                argrepr = f"{preposition} L{lbl}"
            elif deop in (LOAD_FAST_LOAD_FAST, LOAD_FAST_BORROW_LOAD_FAST_BORROW, STORE_FAST_LOAD_FAST, STORE_FAST_STORE_FAST):
                arg1 = arg >> 4
                arg2 = arg & 15
                val1, argrepr1 = _get_name_info(arg1, self.varname_from_oparg)
                val2, argrepr2 = _get_name_info(arg2, self.varname_from_oparg)
                argrepr = argrepr1 + ", " + argrepr2
                argval = val1, val2
            elif deop in haslocal or deop in hasfree:
                argval, argrepr = _get_name_info(arg, self.varname_from_oparg)
            elif deop in hascompare:
                argval = cmp_op[arg >> 5]
                argrepr = argval
                if arg & 16:
                    argrepr = f"bool({argrepr})"
            elif deop == CONVERT_VALUE:
                argval = (None, str, repr, ascii)[arg]
                argrepr = ('', 'str', 'repr', 'ascii')[arg]
            elif deop == SET_FUNCTION_ATTRIBUTE:
                argrepr = ', '.join(s for i, s in enumerate(FUNCTION_ATTR_FLAGS)
                                    if arg & (1<<i))
            elif deop == BINARY_OP:
                _, argrepr = _nb_ops[arg]
            elif deop == CALL_INTRINSIC_1:
                argrepr = _intrinsic_1_descs[arg]
            elif deop == CALL_INTRINSIC_2:
                argrepr = _intrinsic_2_descs[arg]
            elif deop == LOAD_COMMON_CONSTANT:
                obj = _common_constants[arg]
                if isinstance(obj, type):
                    argrepr = obj.__name__
                else:
                    argrepr = repr(obj)
            elif deop == LOAD_SPECIAL:
                argrepr = _special_method_names[arg]
            elif deop == IS_OP:
                argrepr = 'is not' if argval else 'is'
            elif deop == CONTAINS_OP:
                argrepr = 'not in' if argval else 'in'
        return argval, argrepr

def get_instructions(x, *, first_line=None, show_caches=None, adaptive=False):
    """Iterator for the opcodes in methods, functions or code

    Generates a series of Instruction named tuples giving the details of
    each operations in the supplied code.

    If *first_line* is not None, it indicates the line number that should
    be reported for the first source line in the disassembled code.
    Otherwise, the source line information (if any) is taken directly from
    the disassembled code object.
    """
    co = _get_code_object(x)
    linestarts = dict(findlinestarts(co))
    if first_line is not None:
        line_offset = first_line - co.co_firstlineno
    else:
        line_offset = 0

    original_code = co.co_code
    arg_resolver = ArgResolver(co_consts=co.co_consts,
                               names=co.co_names,
                               varname_from_oparg=co._varname_from_oparg,
                               labels_map=_make_labels_map(original_code))
    return _get_instructions_bytes(_get_code_array(co, adaptive),
                                   linestarts=linestarts,
                                   line_offset=line_offset,
                                   co_positions=co.co_positions(),
                                   original_code=original_code,
                                   arg_resolver=arg_resolver)

def _get_const_value(op, arg, co_consts):
    """Helper to get the value of the const in a hasconst op.

       Returns the dereferenced constant if this is possible.
       Otherwise (if it is a LOAD_CONST and co_consts is not
       provided) returns the dis.UNKNOWN sentinel.
    """
    assert op in hasconst or op == LOAD_SMALL_INT

    if op == LOAD_SMALL_INT:
        return arg
    argval = UNKNOWN
    if co_consts is not None:
        argval = co_consts[arg]
    return argval

def _get_const_info(op, arg, co_consts):
    """Helper to get optional details about const references

       Returns the dereferenced constant and its repr if the value
       can be calculated.
       Otherwise returns the sentinel value dis.UNKNOWN for the value
       and an empty string for its repr.
    """
    argval = _get_const_value(op, arg, co_consts)
    argrepr = repr(argval) if argval is not UNKNOWN else ''
    return argval, argrepr

def _get_name_info(name_index, get_name, **extrainfo):
    """Helper to get optional details about named references

       Returns the dereferenced name as both value and repr if the name
       list is defined.
       Otherwise returns the sentinel value dis.UNKNOWN for the value
       and an empty string for its repr.
    """
    if get_name is not None:
        argval = get_name(name_index, **extrainfo)
        return argval, argval
    else:
        return UNKNOWN, ''

def _parse_varint(iterator):
    b = next(iterator)
    val = b & 63
    while b&64:
        val <<= 6
        b = next(iterator)
        val |= b&63
    return val

def _parse_exception_table(code):
    iterator = iter(code.co_exceptiontable)
    entries = []
    try:
        while True:
            start = _parse_varint(iterator)*2
            length = _parse_varint(iterator)*2
            end = start + length
            target = _parse_varint(iterator)*2
            dl = _parse_varint(iterator)
            depth = dl >> 1
            lasti = bool(dl&1)
            entries.append(_ExceptionTableEntry(start, end, target, depth, lasti))
    except StopIteration:
        return entries

def _is_backward_jump(op):
    return opname[op] in ('JUMP_BACKWARD',
                          'JUMP_BACKWARD_NO_INTERRUPT',
                          'END_ASYNC_FOR') # Not really a jump, but it has a "target"

def _get_instructions_bytes(code, linestarts=None, line_offset=0, co_positions=None,
                            original_code=None, arg_resolver=None):
    """Iterate over the instructions in a bytecode string.

    Generates a sequence of Instruction namedtuples giving the details of each
    opcode.

    """
    # Use the basic, unadaptive code for finding labels and actually walking the
    # bytecode, since replacements like ENTER_EXECUTOR and INSTRUMENTED_* can
    # mess that logic up pretty badly:
    original_code = original_code or code
    co_positions = co_positions or iter(())

    starts_line = False
    local_line_number = None
    line_number = None
    for offset, start_offset, op, arg in _unpack_opargs(original_code):
        if linestarts is not None:
            starts_line = offset in linestarts
            if starts_line:
                local_line_number = linestarts[offset]
            if local_line_number is not None:
                line_number = local_line_number + line_offset
            else:
                line_number = None
        positions = Positions(*next(co_positions, ()))
        deop = _deoptop(op)
        op = code[offset]

        if arg_resolver:
            argval, argrepr = arg_resolver.get_argval_argrepr(op, arg, offset)
        else:
            argval, argrepr = arg, repr(arg)

        caches = _get_cache_size(_all_opname[deop])
        # Advance the co_positions iterator:
        for _ in range(caches):
            next(co_positions, ())

        if caches:
            cache_info = []
            cache_offset = offset
            for name, size in _cache_format[opname[deop]].items():
                data = code[cache_offset + 2: cache_offset + 2 + 2 * size]
                cache_offset += size * 2
                cache_info.append((name, size, data))
        else:
            cache_info = None

        label = arg_resolver.get_label_for_offset(offset) if arg_resolver else None
        yield Instruction(_all_opname[op], op, arg, argval, argrepr,
                          offset, start_offset, starts_line, line_number,
                          label, positions, cache_info)


def disassemble(co, lasti=-1, *, file=None, show_caches=False, adaptive=False,
                show_offsets=False, show_positions=False):
    """Disassemble a code object."""
    linestarts = dict(findlinestarts(co))
    exception_entries = _parse_exception_table(co)
    if show_positions:
        lineno_width = _get_positions_width(co)
    else:
        lineno_width = _get_lineno_width(linestarts)
    labels_map = _make_labels_map(co.co_code, exception_entries=exception_entries)
    label_width = 4 + len(str(len(labels_map)))
    formatter = Formatter(file=file,
                          lineno_width=lineno_width,
                          offset_width=len(str(max(len(co.co_code) - 2, 9999))) if show_offsets else 0,
                          label_width=label_width,
                          show_caches=show_caches,
                          show_positions=show_positions)
    arg_resolver = ArgResolver(co_consts=co.co_consts,
                               names=co.co_names,
                               varname_from_oparg=co._varname_from_oparg,
                               labels_map=labels_map)
    _disassemble_bytes(_get_code_array(co, adaptive), lasti, linestarts,
                       exception_entries=exception_entries, co_positions=co.co_positions(),
                       original_code=co.co_code, arg_resolver=arg_resolver, formatter=formatter)

def _disassemble_recursive(co, *, file=None, depth=None, show_caches=False, adaptive=False, show_offsets=False, show_positions=False):
    disassemble(co, file=file, show_caches=show_caches, adaptive=adaptive, show_offsets=show_offsets, show_positions=show_positions)
    if depth is None or depth > 0:
        if depth is not None:
            depth = depth - 1
        for x in co.co_consts:
            if hasattr(x, 'co_code'):
                print(file=file)
                print("Disassembly of %r:" % (x,), file=file)
                _disassemble_recursive(
                    x, file=file, depth=depth, show_caches=show_caches,
                    adaptive=adaptive, show_offsets=show_offsets, show_positions=show_positions
                )


def _make_labels_map(original_code, exception_entries=()):
    jump_targets = set(findlabels(original_code))
    labels = set(jump_targets)
    for start, end, target, _, _ in exception_entries:
        labels.add(start)
        labels.add(end)
        labels.add(target)
    labels = sorted(labels)
    labels_map = {offset: i+1 for (i, offset) in enumerate(sorted(labels))}
    for e in exception_entries:
        e.start_label = labels_map[e.start]
        e.end_label = labels_map[e.end]
        e.target_label = labels_map[e.target]
    return labels_map

_NO_LINENO = '  --'

def _get_lineno_width(linestarts):
    if linestarts is None:
        return 0
    maxlineno = max(filter(None, linestarts.values()), default=-1)
    if maxlineno == -1:
        # Omit the line number column entirely if we have no line number info
        return 0
    lineno_width = max(3, len(str(maxlineno)))
    if lineno_width < len(_NO_LINENO) and None in linestarts.values():
        lineno_width = len(_NO_LINENO)
    return lineno_width

def _get_positions_width(code):
    # Positions are formatted as 'LINE:COL-ENDLINE:ENDCOL ' (note trailing space).
    # A missing component appears as '?', and when all components are None, we
    # render '_NO_LINENO'. thus the minimum width is 1 + len(_NO_LINENO).
    #
    # If all values are missing, positions are not printed (i.e. positions_width = 0).
    has_value = False
    values_width = 0
    for positions in code.co_positions():
        has_value |= any(isinstance(p, int) for p in positions)
        width = sum(1 if p is None else len(str(p)) for p in positions)
        values_width = max(width, values_width)
    if has_value:
        # 3 = number of separators in a normal format
        return 1 + max(len(_NO_LINENO), 3 + values_width)
    return 0

def _disassemble_bytes(code, lasti=-1, linestarts=None,
                       *, line_offset=0, exception_entries=(),
                       co_positions=None, original_code=None,
                       arg_resolver=None, formatter=None):

    assert formatter is not None
    assert arg_resolver is not None

    instrs = _get_instructions_bytes(code, linestarts=linestarts,
                                           line_offset=line_offset,
                                           co_positions=co_positions,
                                           original_code=original_code,
                                           arg_resolver=arg_resolver)

    print_instructions(instrs, exception_entries, formatter, lasti=lasti)


def print_instructions(instrs, exception_entries, formatter, lasti=-1):
    for instr in instrs:
        # Each CACHE takes 2 bytes
        is_current_instr = instr.offset <= lasti \
            <= instr.offset + 2 * _get_cache_size(_all_opname[_deoptop(instr.opcode)])
        formatter.print_instruction(instr, is_current_instr)

    formatter.print_exception_table(exception_entries)

def _disassemble_str(source, **kwargs):
    """Compile the source string, then disassemble the code object."""
    _disassemble_recursive(_try_compile(source, '<dis>'), **kwargs)

disco = disassemble                     # XXX For backwards compatibility


# Rely on C `int` being 32 bits for oparg
_INT_BITS = 32
# Value for c int when it overflows
_INT_OVERFLOW = 2 ** (_INT_BITS - 1)

def _unpack_opargs(code):
    extended_arg = 0
    extended_args_offset = 0  # Number of EXTENDED_ARG instructions preceding the current instruction
    caches = 0
    for i in range(0, len(code), 2):
        # Skip inline CACHE entries:
        if caches:
            caches -= 1
            continue
        op = code[i]
        deop = _deoptop(op)
        caches = _get_cache_size(_all_opname[deop])
        if deop in hasarg:
            arg = code[i+1] | extended_arg
            extended_arg = (arg << 8) if deop == EXTENDED_ARG else 0
            # The oparg is stored as a signed integer
            # If the value exceeds its upper limit, it will overflow and wrap
            # to a negative integer
            if extended_arg >= _INT_OVERFLOW:
                extended_arg -= 2 * _INT_OVERFLOW
        else:
            arg = None
            extended_arg = 0
        if deop == EXTENDED_ARG:
            extended_args_offset += 1
            yield (i, i, op, arg)
        else:
            start_offset = i - extended_args_offset*2
            yield (i, start_offset, op, arg)
            extended_args_offset = 0

def findlabels(code):
    """Detect all offsets in a byte code which are jump targets.

    Return the list of offsets.

    """
    labels = []
    for offset, _, op, arg in _unpack_opargs(code):
        if arg is not None:
            label = _get_jump_target(op, arg, offset)
            if label is None:
                continue
            if label not in labels:
                labels.append(label)
    return labels

def findlinestarts(code):
    """Find the offsets in a byte code which are start of lines in the source.

    Generate pairs (offset, lineno)
    lineno will be an integer or None the offset does not have a source line.
    """

    lastline = False # None is a valid line number
    for start, end, line in code.co_lines():
        if line is not lastline:
            lastline = line
            yield start, line
    return

def _find_imports(co):
    """Find import statements in the code

    Generate triplets (name, level, fromlist) where
    name is the imported module and level, fromlist are
    the corresponding args to __import__.
    """
    IMPORT_NAME = opmap['IMPORT_NAME']

    consts = co.co_consts
    names = co.co_names
    opargs = [(op, arg) for _, _, op, arg in _unpack_opargs(co.co_code)
                  if op != EXTENDED_ARG]
    for i, (op, oparg) in enumerate(opargs):
        if op == IMPORT_NAME and i >= 2:
            from_op = opargs[i-1]
            level_op = opargs[i-2]
            if (from_op[0] in hasconst and
                (level_op[0] in hasconst or level_op[0] == LOAD_SMALL_INT)):
                level = _get_const_value(level_op[0], level_op[1], consts)
                fromlist = _get_const_value(from_op[0], from_op[1], consts)
                yield (names[oparg], level, fromlist)

def _find_store_names(co):
    """Find names of variables which are written in the code

    Generate sequence of strings
    """
    STORE_OPS = {
        opmap['STORE_NAME'],
        opmap['STORE_GLOBAL']
    }

    names = co.co_names
    for _, _, op, arg in _unpack_opargs(co.co_code):
        if op in STORE_OPS:
            yield names[arg]


class Bytecode:
    """The bytecode operations of a piece of code

    Instantiate this with a function, method, other compiled object, string of
    code, or a code object (as returned by compile()).

    Iterating over this yields the bytecode operations as Instruction instances.
    """
    def __init__(self, x, *, first_line=None, current_offset=None, show_caches=False, adaptive=False, show_offsets=False, show_positions=False):
        self.codeobj = co = _get_code_object(x)
        if first_line is None:
            self.first_line = co.co_firstlineno
            self._line_offset = 0
        else:
            self.first_line = first_line
            self._line_offset = first_line - co.co_firstlineno
        self._linestarts = dict(findlinestarts(co))
        self._original_object = x
        self.current_offset = current_offset
        self.exception_entries = _parse_exception_table(co)
        self.show_caches = show_caches
        self.adaptive = adaptive
        self.show_offsets = show_offsets
        self.show_positions = show_positions

    def __iter__(self):
        co = self.codeobj
        original_code = co.co_code
        labels_map = _make_labels_map(original_code, self.exception_entries)
        arg_resolver = ArgResolver(co_consts=co.co_consts,
                                   names=co.co_names,
                                   varname_from_oparg=co._varname_from_oparg,
                                   labels_map=labels_map)
        return _get_instructions_bytes(_get_code_array(co, self.adaptive),
                                       linestarts=self._linestarts,
                                       line_offset=self._line_offset,
                                       co_positions=co.co_positions(),
                                       original_code=original_code,
                                       arg_resolver=arg_resolver)

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__,
                                 self._original_object)

    @classmethod
    def from_traceback(cls, tb, *, show_caches=False, adaptive=False):
        """ Construct a Bytecode from the given traceback """
        while tb.tb_next:
            tb = tb.tb_next
        return cls(
            tb.tb_frame.f_code, current_offset=tb.tb_lasti, show_caches=show_caches, adaptive=adaptive
        )

    def info(self):
        """Return formatted information about the code object."""
        return _format_code_info(self.codeobj)

    def dis(self):
        """Return a formatted view of the bytecode operations."""
        co = self.codeobj
        if self.current_offset is not None:
            offset = self.current_offset
        else:
            offset = -1
        with io.StringIO() as output:
            code = _get_code_array(co, self.adaptive)
            offset_width = len(str(max(len(code) - 2, 9999))) if self.show_offsets else 0
            if self.show_positions:
                lineno_width = _get_positions_width(co)
            else:
                lineno_width = _get_lineno_width(self._linestarts)
            labels_map = _make_labels_map(co.co_code, self.exception_entries)
            label_width = 4 + len(str(len(labels_map)))
            formatter = Formatter(file=output,
                                  lineno_width=lineno_width,
                                  offset_width=offset_width,
                                  label_width=label_width,
                                  line_offset=self._line_offset,
                                  show_caches=self.show_caches,
                                  show_positions=self.show_positions)

            arg_resolver = ArgResolver(co_consts=co.co_consts,
                                       names=co.co_names,
                                       varname_from_oparg=co._varname_from_oparg,
                                       labels_map=labels_map)
            _disassemble_bytes(code,
                               linestarts=self._linestarts,
                               line_offset=self._line_offset,
                               lasti=offset,
                               exception_entries=self.exception_entries,
                               co_positions=co.co_positions(),
                               original_code=co.co_code,
                               arg_resolver=arg_resolver,
                               formatter=formatter)
            return output.getvalue()


def main(args=None):
    import argparse

    parser = argparse.ArgumentParser(color=True)
    parser.add_argument('-C', '--show-caches', action='store_true',
                        help='show inline caches')
    parser.add_argument('-O', '--show-offsets', action='store_true',
                        help='show instruction offsets')
    parser.add_argument('-P', '--show-positions', action='store_true',
                        help='show instruction positions')
    parser.add_argument('-S', '--specialized', action='store_true',
                        help='show specialized bytecode')
    parser.add_argument('infile', nargs='?', default='-')
    args = parser.parse_args(args=args)
    if args.infile == '-':
        name = '<stdin>'
        source = sys.stdin.buffer.read()
    else:
        name = args.infile
        with open(args.infile, 'rb') as infile:
            source = infile.read()
    code = compile(source, name, "exec")
    dis(code, show_caches=args.show_caches, adaptive=args.specialized,
        show_offsets=args.show_offsets, show_positions=args.show_positions)

if __name__ == "__main__":
    main()



from __future__ import annotations

import os
import re
import shlex
import shutil
import struct
import subprocess
import sys
import textwrap
from collections.abc import Iterable
from typing import TYPE_CHECKING, TypedDict

from ._importlib import metadata, resources

if TYPE_CHECKING:
    from typing_extensions import Self

from .warnings import SetuptoolsWarning

from distutils.command.build_scripts import first_line_re
from distutils.util import get_platform


class _SplitArgs(TypedDict, total=False):
    comments: bool
    posix: bool


class CommandSpec(list):
    """
    A command spec for a #! header, specified as a list of arguments akin to
    those passed to Popen.
    """

    options: list[str] = []
    split_args = _SplitArgs()

    @classmethod
    def best(cls):
        """
        Choose the best CommandSpec class based on environmental conditions.
        """
        return cls

    @classmethod
    def _sys_executable(cls):
        _default = os.path.normpath(sys.executable)
        return os.environ.get('__PYVENV_LAUNCHER__', _default)

    @classmethod
    def from_param(cls, param: Self | str | Iterable[str] | None) -> Self:
        """
        Construct a CommandSpec from a parameter to build_scripts, which may
        be None.
        """
        if isinstance(param, cls):
            return param
        if isinstance(param, str):
            return cls.from_string(param)
        if isinstance(param, Iterable):
            return cls(param)
        if param is None:
            return cls.from_environment()
        raise TypeError(f"Argument has an unsupported type {type(param)}")

    @classmethod
    def from_environment(cls):
        return cls([cls._sys_executable()])

    @classmethod
    def from_string(cls, string: str) -> Self:
        """
        Construct a command spec from a simple string representing a command
        line parseable by shlex.split.
        """
        items = shlex.split(string, **cls.split_args)
        return cls(items)

    def install_options(self, script_text: str):
        self.options = shlex.split(self._extract_options(script_text))
        cmdline = subprocess.list2cmdline(self)
        if not isascii(cmdline):
            self.options[:0] = ['-x']

    @staticmethod
    def _extract_options(orig_script):
        """
        Extract any options from the first line of the script.
        """
        first = (orig_script + '\n').splitlines()[0]
        match = _first_line_re().match(first)
        options = match.group(1) or '' if match else ''
        return options.strip()

    def as_header(self):
        return self._render(self + list(self.options))

    @staticmethod
    def _strip_quotes(item):
        _QUOTES = '"\''
        for q in _QUOTES:
            if item.startswith(q) and item.endswith(q):
                return item[1:-1]
        return item

    @staticmethod
    def _render(items):
        cmdline = subprocess.list2cmdline(
            CommandSpec._strip_quotes(item.strip()) for item in items
        )
        return '#!' + cmdline + '\n'


class WindowsCommandSpec(CommandSpec):
    split_args = _SplitArgs(posix=False)


class ScriptWriter:
    """
    Encapsulates behavior around writing entry point scripts for console and
    gui apps.
    """

    template = textwrap.dedent(
        r"""
        # EASY-INSTALL-ENTRY-SCRIPT: %(spec)r,%(group)r,%(name)r
        import re
        import sys

        # for compatibility with easy_install; see #2198
        __requires__ = %(spec)r

        try:
            from importlib.metadata import distribution
        except ImportError:
            try:
                from importlib_metadata import distribution
            except ImportError:
                from pkg_resources import load_entry_point


        def importlib_load_entry_point(spec, group, name):
            dist_name, _, _ = spec.partition('==')
            matches = (
                entry_point
                for entry_point in distribution(dist_name).entry_points
                if entry_point.group == group and entry_point.name == name
            )
            return next(matches).load()


        globals().setdefault('load_entry_point', importlib_load_entry_point)


        if __name__ == '__main__':
            sys.argv[0] = re.sub(r'(-script\.pyw?|\.exe)?$', '', sys.argv[0])
            sys.exit(load_entry_point(%(spec)r, %(group)r, %(name)r)())
        """
    ).lstrip()

    command_spec_class = CommandSpec

    @classmethod
    def get_args(cls, dist, header=None):
        """
        Yield write_script() argument tuples for a distribution's
        console_scripts and gui_scripts entry points.
        """

        # If distribution is not an importlib.metadata.Distribution, assume
        # it's a pkg_resources.Distribution and transform it.
        if not hasattr(dist, 'entry_points'):
            SetuptoolsWarning.emit("Unsupported distribution encountered.")
            dist = metadata.Distribution.at(dist.egg_info)

        if header is None:
            header = cls.get_header()
        spec = f'{dist.name}=={dist.version}'
        for type_ in 'console', 'gui':
            group = f'{type_}_scripts'
            for ep in dist.entry_points.select(group=group):
                name = ep.name
                cls._ensure_safe_name(ep.name)
                script_text = cls.template % locals()
                args = cls._get_script_args(type_, ep.name, header, script_text)
                yield from args

    @staticmethod
    def _ensure_safe_name(name):
        """
        Prevent paths in *_scripts entry point names.
        """
        has_path_sep = re.search(r'[\\/]', name)
        if has_path_sep:
            raise ValueError("Path separators not allowed in script names")

    @classmethod
    def best(cls):
        """
        Select the best ScriptWriter for this environment.
        """
        if sys.platform == 'win32' or (os.name == 'java' and os._name == 'nt'):
            return WindowsScriptWriter.best()
        else:
            return cls

    @classmethod
    def _get_script_args(cls, type_, name, header, script_text):
        # Simply write the stub with no extension.
        yield (name, header + script_text)

    @classmethod
    def get_header(
        cls,
        script_text: str = "",
        executable: str | CommandSpec | Iterable[str] | None = None,
    ) -> str:
        """Create a #! line, getting options (if any) from script_text"""
        cmd = cls.command_spec_class.best().from_param(executable)
        cmd.install_options(script_text)
        return cmd.as_header()


class WindowsScriptWriter(ScriptWriter):
    command_spec_class = WindowsCommandSpec

    @classmethod
    def best(cls):
        """
        Select the best ScriptWriter suitable for Windows
        """
        writer_lookup = dict(
            executable=WindowsExecutableLauncherWriter,
            natural=cls,
        )
        # for compatibility, use the executable launcher by default
        launcher = os.environ.get('SETUPTOOLS_LAUNCHER', 'executable')
        return writer_lookup[launcher]

    @classmethod
    def _get_script_args(cls, type_, name, header, script_text):
        "For Windows, add a .py extension"
        ext = dict(console='.pya', gui='.pyw')[type_]
        if ext not in os.environ['PATHEXT'].lower().split(';'):
            msg = (
                "{ext} not listed in PATHEXT; scripts will not be "
                "recognized as executables."
            ).format(**locals())
            SetuptoolsWarning.emit(msg)
        old = ['.pya', '.py', '-script.py', '.pyc', '.pyo', '.pyw', '.exe']
        old.remove(ext)
        header = cls._adjust_header(type_, header)
        blockers = [name + x for x in old]
        yield name + ext, header + script_text, 't', blockers

    @classmethod
    def _adjust_header(cls, type_, orig_header):
        """
        Make sure 'pythonw' is used for gui and 'python' is used for
        console (regardless of what sys.executable is).
        """
        pattern = 'pythonw.exe'
        repl = 'python.exe'
        if type_ == 'gui':
            pattern, repl = repl, pattern
        pattern_ob = re.compile(re.escape(pattern), re.IGNORECASE)
        new_header = pattern_ob.sub(string=orig_header, repl=repl)
        return new_header if cls._use_header(new_header) else orig_header

    @staticmethod
    def _use_header(new_header):
        """
        Should _adjust_header use the replaced header?

        On non-windows systems, always use. On
        Windows systems, only use the replaced header if it resolves
        to an executable on the system.
        """
        clean_header = new_header[2:-1].strip('"')
        return sys.platform != 'win32' or shutil.which(clean_header)


class WindowsExecutableLauncherWriter(WindowsScriptWriter):
    @classmethod
    def _get_script_args(cls, type_, name, header, script_text):
        """
        For Windows, add a .py extension and an .exe launcher
        """
        if type_ == 'gui':
            launcher_type = 'gui'
            ext = '-script.pyw'
            old = ['.pyw']
        else:
            launcher_type = 'cli'
            ext = '-script.py'
            old = ['.py', '.pyc', '.pyo']
        hdr = cls._adjust_header(type_, header)
        blockers = [name + x for x in old]
        yield (name + ext, hdr + script_text, 't', blockers)
        yield (
            name + '.exe',
            get_win_launcher(launcher_type),
            'b',  # write in binary mode
        )
        if not is_64bit():
            # install a manifest for the launcher to prevent Windows
            # from detecting it as an installer (which it will for
            #  launchers like easy_install.exe). Consider only
            #  adding a manifest for launchers detected as installers.
            #  See Distribute #143 for details.
            m_name = name + '.exe.manifest'
            yield (m_name, load_launcher_manifest(name), 't')


def get_win_launcher(type):
    """
    Load the Windows launcher (executable) suitable for launching a script.

    `type` should be either 'cli' or 'gui'

    Returns the executable as a byte string.
    """
    launcher_fn = f'{type}.exe'
    if is_64bit():
        if get_platform() == "win-arm64":
            launcher_fn = launcher_fn.replace(".", "-arm64.")
        else:
            launcher_fn = launcher_fn.replace(".", "-64.")
    else:
        launcher_fn = launcher_fn.replace(".", "-32.")
    return resources.files('setuptools').joinpath(launcher_fn).read_bytes()


def load_launcher_manifest(name):
    res = resources.files(__name__).joinpath('launcher manifest.xml')
    return res.read_text(encoding='utf-8') % vars()


def _first_line_re():
    """
    Return a regular expression based on first_line_re suitable for matching
    strings.
    """
    if isinstance(first_line_re.pattern, str):
        return first_line_re

    # first_line_re in Python >=3.1.4 and >=3.2.1 is a bytes pattern.
    return re.compile(first_line_re.pattern.decode())


def is_64bit():
    return struct.calcsize("P") == 8


def isascii(s):
    try:
        s.encode('ascii')
    except UnicodeError:
        return False
    return True
